#! /usr/bin/env perl
#
# tag: ewok install apache config files installation script
#

use Getopt::Std;
use File::Basename;
use File::Path;
use File::Copy;
use File::Compare;
use strict;

my $build_dir = dirname($0);
my $conf_dir = "$build_dir/../conf";
my $cachefile = "$build_dir/config.cache";
my @conf = qw(ewok.conf ewok-startup.pl);
for (@conf) {
    die "configuration file '$conf_dir/$_' not found" if ! -f "$conf_dir/$_";
}
my $hr = '-' x 78;

# Parse options
my %opts;
getopts('n', \%opts);
my $noop = $opts{n};

# Load the cache
die "cannot find cache file $cachefile - have you run configure.pl?" 
    if ! -f $cachefile;
my $cache;
open CACHE, "<$cachefile"
    or die "cannot open '$cachefile' for read";
{
    local $/;
    eval <CACHE>;
}
close CACHE;

# Make apache_conf dir if doesn't exist
if (! -d $cache->{apache_conf}) {
    print "mkdir $cache->{apache_conf}\n";
    if (! $noop) {
        my @new = mkpath($cache->{apache_conf});
        die "creating configuration directory $cache->{apache_conf} failed: $!"
            unless $new[$#new] eq $cache->{apache_conf};
        chown $cache->{httpd_uid}, -1, $cache->{apache_conf}
            if $> == 0 && $cache->{httpd_uid};
    }
}

# Copy files
my @existing = ();
for (@conf) {
    my $filename = $_;
    my $copy = 1;
    if (-f "$cache->{apache_conf}/$_") {
        if (compare("conf/$_", "$cache->{apache_conf}/$_") == 0) {
            print "-> $cache->{apache_conf}/$_ identical (skipping)\n";
            $copy = 0;
        }
        else {
            print "-> Warning: $cache->{apache_conf}/$_ already exists - keeping existing version\n";
            push @existing, $filename;
            $filename .= '.new';
        }
    }
    print "cp conf/$_ $cache->{apache_conf}/$filename\n" if $copy;
    if (! $noop) {
        if ($copy) {
            copy("$conf_dir/$_", "$cache->{apache_conf}/$filename");
            chown $cache->{httpd_uid}, -1, "$cache->{apache_conf}/$filename"
                if $> == 0 && $cache->{httpd_uid};
        }
        # Create symlink in ewok_home/conf (reset filename to initial)
        $filename = $_;
        if (! -e "$cache->{ewok_home}/conf") {
            mkpath "$cache->{ewok_home}/conf" 
                or die "mkdir '$cache->{ewok_home}/conf' failed: $!";
            chown $cache->{httpd_uid}, -1, "$cache->{ewok_home}/conf"
                if $> == 0 && $cache->{httpd_uid};
        }
        my $link = "$cache->{ewok_home}/conf/$filename";
        if (! -e $link) {
            symlink "$cache->{apache_conf}/$filename", $link
                or die "symlink $cache->{apache_conf}/me, $link failed: $!";
            chown $cache->{httpd_uid}, -1, $link
                if $> == 0 && $cache->{httpd_uid};
        }
        elsif (! -l $link) {
            print "-> Warning: unable to create $cache->{ewok_home}/conf/$filename\nlink - non-link already exists\n";
        }
    }
}

# Modify httpd.conf, if it exists
my $httpd = "$cache->{apache_conf}/httpd-perl.conf";
$httpd = "$cache->{apache_conf}/httpd.conf" if ! -f $httpd;
if (-f $httpd && -w $httpd) {
    open HTTPD, "<$httpd"
        or die "cannot open '$httpd' for read: $!";
    my $hconf;
    {
        local $/;
        $hconf = <HTTPD>;
    }
    close HTTPD;
    # Add the include for "ewok.conf" unless it's already present
    if ($hconf && $hconf !~ m/^(#\s*)?Include.*\bewok.conf"/m) {
        print "-> adding ewok.conf include directive to $httpd\n";
        if (! $noop) {
            open HTTPD, ">>$httpd"
                or die "cannot open '$httpd' for append $!";
            print HTTPD <<EOD;

# Include ewok configuration directives
Include "conf/ewok.conf"
EOD
            close HTTPD;
        }
    }
    elsif ($1) {
        print "-> Warning: $httpd contains ewok.conf include directive,\n   but it's commented out - you'll need to uncomment it to enable ewok\n";
    }
    else {
        print "-> $httpd already contains ewok.conf include directive\n";
    }
}
else {
    print <<EOD;
$hr
ewok was unable to locate your main apache config file (usually httpd.conf) to
add the appropriate include directive for ewok. Please add the following line
to your apache config file so that apache can find ewok:
    Include "conf/ewok.conf"
EOD
}

# If config files already existed, warn the user 
if (@existing) {
    print <<EOD;
$hr
The following existing ewok config files are different to their distributed 
versions:
EOD
    print '    ' . join("\n    ", @existing), "\n";
    print <<EOD;
ewok has preserved your existing config files and installed its new versions
with a '.new' suffix. You should probably compare the old and new versions
to confirm that the only differences are your site-local customizations. If
you do find any non-local differences, you should probably copy them into 
your existing configs. For example:
    \$ cd $cache->{apache_conf}
EOD
    print "    \$ diff $_ $_.new | more\n" foreach @existing;
}

# What's next
print <<EOD;
$hr
ewok has successfully installed. If you haven't previously configured ewok,
you should now setup an initial staging directory, modify the MANDATORY
section in EWOK_HOME/conf/Config.pm, and add your staging directory to
EWOK_HOME/conf/ewok.conf. See the INSTALL file's "Minimal Configuration
Survival Guide" for details. Have fun!
$hr
EOD
