# tag: ewok authorisation module

package Ewok::Authz;

require Exporter;
use Data::Dumper;
use HTML::EmbperlObject;
use Ewok::Config qw(:all);
use Ewok::Util qw(:load &hashInvert);

use strict;
use vars qw(
    @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS 
    $VIEW_PRIV $VIEWPUBL_PRIV $EDIT_PRIV $PUBL_PRIV $AUTHZ_PRIV $TMPL_PRIV 
    $ADMIN_PRIV @PRIV @APRIV @privHierarchy %privlist
);

@ISA = qw(Exporter);
@EXPORT = ();
@EXPORT_OK = qw(
    &authzMatch
    &authzGetUserHash
    &authzAdmin
    $VIEW_PRIV
    $VIEWPUBL_PRIV
    $EDIT_PRIV
    $PUBL_PRIV
    $TMPL_PRIV
    $AUTHZ_PRIV
    $ADMIN_PRIV
    @PRIV
    @APRIV
    @privHierarchy
    &authzGetAll
    &authzSave
    &authzObjCheck
    &squashPriv
);
%EXPORT_TAGS = ( 
    common => [qw(&authzMatch &authzGetUserHash &authzAdmin
        $VIEW_PRIV $VIEWPUBL_PRIV $EDIT_PRIV $PUBL_PRIV 
        $AUTHZ_PRIV $TMPL_PRIV $ADMIN_PRIV)],
    all =>    [@EXPORT_OK],
);

# Constants
BEGIN {
    $VIEW_PRIV = 'view';
    $VIEWPUBL_PRIV = 'viewpubl';
    $EDIT_PRIV = 'edit';
    $PUBL_PRIV = 'publish';
    $AUTHZ_PRIV = 'permissions';
    $TMPL_PRIV = 'templates';
    $ADMIN_PRIV = 'admin';

    # Collections
    @PRIV = ($VIEW_PRIV, $VIEWPUBL_PRIV, $EDIT_PRIV, $PUBL_PRIV, $AUTHZ_PRIV, $TMPL_PRIV);
    @APRIV = (@PRIV, $ADMIN_PRIV);
    # The first four privileges (view, edit, publ, authz) are hierarchical, each
    #   including all the ones before
    @privHierarchy = ($AUTHZ_PRIV, $PUBL_PRIV, $EDIT_PRIV, $VIEW_PRIV);
    %privlist = (
        $VIEW_PRIV => [ $VIEW_PRIV ],
        $EDIT_PRIV => [ $VIEW_PRIV, $EDIT_PRIV ],
        $PUBL_PRIV => [ $VIEW_PRIV, $EDIT_PRIV, $PUBL_PRIV ],
        $AUTHZ_PRIV => [ $VIEW_PRIV, $EDIT_PRIV, $PUBL_PRIV, $AUTHZ_PRIV ],
        $TMPL_PRIV => [ $TMPL_PRIV ],
        $ADMIN_PRIV => [ @APRIV ],
    );
}

#
# Return the authz hashref applicable to $fs_file, or undef on failure.
#
sub authzGetAll
{
    my ($req, $fs_file) = @_;
    return fileLoad($EWOK_AUTHZ_OBJECT, $fs_file, 
        { inherit => 1, eval => 1, name => 'authz', notfound => {}, 
          addpath => undef, });
}

#
# Returns true if $user (or $req->{user}) is an admin, false otherwise.
#
sub authzAdmin
{
    my $req = shift;
    my $user = shift || $req->{user};
    return scalar grep { $user eq $_ } @EWOK_ADMINS;
}

#
# Return the authz value for $user from the authz file applicable
#   to $fs_file. Return 0 if $user entry does not exist, or undef
#   for any other error.
#
sub authzGetUser
{
    my $req = shift;
    my $fs_file = shift;
    my $user = shift;
    $user ||= $req->{user};
    return undef if ! $user;

    my $authz = authzGetAll($req, $fs_file);
    return undef if ref $authz ne 'HASH';
    return 0 if not exists $authz->{$user};

    return $authz->{$user};
}

#
# Return true if the authz entry for $fs_file for $user includes $priv.
#   Calls authzGetUser to get the entry in question.
#
sub authzMatch
{
    my $req = shift;
    my $fs_file = shift;
    my $priv = shift;
    my $user = shift || $req->{user};
    return undef if ! $user;

    # Check if valid privilege
    return undef if ! grep /^$priv$/, @APRIV;

    # Check if in @EWOK_ADMINS
    return 2 if grep { $user eq $_ } @EWOK_ADMINS;

    # Avoid 'admin' spoofing in authz.epl
    return undef if $priv eq $ADMIN_PRIV;

    # $VIEW_PRIV is implied by any of the others - sufficient to find anything
    $priv = '\\w+' if $priv eq $VIEW_PRIV;

    # Check the user entry
    my $userEntry = authzGetUser($req, $fs_file, $user);
    return undef if ! defined $userEntry;
    my @userPriv = split /\s+/, $userEntry;
    for my $userPriv (@userPriv) {
        return 1 if grep /^$priv$/, @{$privlist{$userPriv}};
    }

    # Check the 'ALL' entry
    my $allEntry = authzGetUser($req, $fs_file, 'ALL');
    return undef if ! defined $userEntry;
    my @allPriv = split /\s+/, $allEntry;
    for my $allPriv (@allPriv) {
        return 1 if grep /^$priv$/, @{$privlist{$allPriv}};
    }

    return 0;
}

#
# Return a privilege hashref for $user from the authz file applicable
#   to $fs_file, or undef on failure.
#
sub authzGetUserHash
{
    my ($req, $fs_file, $user) = @_;
    $user ||= $req->{user};
    my $authz = {};

    # Special handling for EWOK_ADMINS - give all privileges and return
    if (authzAdmin($req, $user)) {
        $authz = { map { $_ => 1 } @PRIV };
        $authz->{$ADMIN_PRIV} = 1;
        $authz->{PRIV} = $ADMIN_PRIV;
        return $authz;
    }

    # Get privs for the user and for 'ALL'
    my $userEntry = authzGetUser(@_);
    $authz->{PRIV} = $userEntry;
    my @userPriv = split /\s+/, $userEntry;
    my $allEntry = authzGetUser($req, $fs_file, 'ALL');
    $authz->{PRIV} ||= $allEntry;
    my @allPriv = split /\s+/, $allEntry;
    for my $priv (@userPriv, @allPriv) {
        if (!exists $authz->{$priv}) {
            # Add the privlist entries for this $priv to $authz
            map { $authz->{$_} = 1 } @{$privlist{$priv}};
        }
    }
    $authz->{PRIV} =~ s/$VIEWPUBL_PRIV\s*//;

    # Catch 'admin' spoofing in authz.epl
    delete $authz->{$ADMIN_PRIV} if exists $authz->{$ADMIN_PRIV};

    return $authz;
}

#
# Check the $authz privileges for all *.epl files in $file (scalar or arrayref).
#   Returns true on success (authorised) or false on failure.
#
sub authzObjCheck
{
    my $file = shift;
    my $authz = shift;
    my $op = shift;
    my $errors = shift;
    my $errCount = scalar @$errors;
    $errors = [] if ref $errors ne 'ARRAY';
    warning("Error: bad authz argument to authzObjCheck.", $errors)
        if ref $authz ne 'HASH';
    return undef if scalar @$errors > $errCount;

    my @filelist = ((ref $file eq 'ARRAY') ? @$file : ( $file ));
    for my $filename (@filelist) {
        next if $filename !~ /\.epl\s*$/o;
        $filename =~ s/^\s*(.*?)\s*$/$1/o;
        # Remove leading username if paste
        $filename =~ s/^.*-([^\-]+\.epl)$/$1/o if $op eq 'paste';
        # if (grep /^$filename$/, @Ewok::Config::EWOK_ADMIN_OBJECTS) {
        if ($EWOK_ADMIN_OBJECTS{$filename}) {
            info("Error: you are not authorised to \l$op $filename.", $errors)
                if ! defined $authz->{$ADMIN_PRIV};
        }
        elsif ($filename eq $Ewok::Config::EWOK_AUTHZ_OBJECT) {
            info("Error: you are not authorised to \l$op $filename.", $errors)
                if ! defined $authz->{$AUTHZ_PRIV};
        }
        elsif (grep /^$filename$/, @Ewok::Config::EWOK_TEMPLATE_OBJECTS) {
            info("Error: you are not authorised to \l$op $filename.", $errors)
                if ! defined $authz->{$TMPL_PRIV};
        }
    }
    return (scalar @$errors <= $errCount);
}

#
# Check and remove implied permissions from the $authz hash
#
sub squashPriv
{
    my $authz = shift;

    # Check the authz hash for implied values
    my ($user, $privstr, @priv);
    while (($user, $privstr) = each %$authz) {
        @priv = split /\s+/, $privstr;
        next if scalar @priv <= 1;
        for (my $i = 0; $i <= $#privHierarchy-1; $i++) {
            if (grep /^$privHierarchy[$i]$/, @priv) {
                # Remove all other items in the privHierarchy
                for (my $j = $i+1; $j <= $#privHierarchy; $j++) {
                    @priv = grep !/^$privHierarchy[$j]$/, @priv;
                }
                last;
            }
        }
        $authz->{$user} = join(' ', @priv);
    }
}

#
# Save the given authz hashref into the authz file within the given directory.
#   Also defines or deletes a .htaccess file in that directory, depending
#   on whether ALL have view privilege in the authz hashref.
# Returns undef on failure, true on success.
#
sub authzSave
{
    my ($req, $dirname, $authz, $arg) = @_;
    die "[authzSave] invalid directory $dirname"
        if $dirname eq '' || index($dirname,'/') < 0;
    die "[authzSave] invalid permissions '$authz'"
        if ref $authz ne 'HASH';
    die "[authzSave] invalid arguments '$arg'"
        if $arg && ref $arg ne 'HASH';
    my $publish = $arg->{publish} || '';
    my $label = $arg->{label} || '';

    # Make sure dirname is a directory
    if (! -d $dirname) {
        $dirname =~ s|/[^/]+$||o;
    }
    else {
        $dirname = substr($dirname, 0, -1) if substr($dirname, -1) eq '/';
    }
    die "[authzSave] '$dirname' is not a directory"
        if ! -d $dirname;

    # Remove implied permissions from the authz hash
    squashPriv($authz);

    # If $publish, demote privileges (edit -> viewpubl, publish -> edit)
    if ($publish) {
        for my $user (keys %$authz) {
            # Map viewpubl-$publish to viewpubl and remove all viewpubl-other
            $authz->{$user} =~ s/$VIEWPUBL_PRIV-$publish/$VIEWPUBL_PRIV/;
            $authz->{$user} =~ s/$VIEWPUBL_PRIV-\w+/ /go;
            
            # Ignore ALL/NONE (GMC: is this right?!!!)
            next if $user eq 'ALL' || $user eq 'NONE';
            # Demote privileges (no 'publish' once published!)
            $authz->{$user} =~ s/\b$EDIT_PRIV\b/$VIEWPUBL_PRIV/o;
            $authz->{$user} =~ s/\b$PUBL_PRIV|$AUTHZ_PRIV\b/$EDIT_PRIV/o;
        }
    }

    # Store the authz hash
    open AUTHZ, ">", "$dirname/$EWOK_AUTHZ_OBJECT" 
        or die "[authzSave] unable to open $dirname/$EWOK_AUTHZ_OBJECT for write: $!";
    $Data::Dumper::Indent = 1;
    print AUTHZ Data::Dumper->Dump([$authz], [qw(authz)]);
    close AUTHZ;
    chmod 0600, "$dirname/$EWOK_AUTHZ_OBJECT";

    # Ignore only $ignore_priv in setting up .htaccess
    my $ignore_priv = $publish ? $VIEW_PRIV : $VIEWPUBL_PRIV;

    # Any ALL privilege except $ignore_priv means no .htaccess is required
    my $all = exists $authz->{ALL} && $authz->{ALL} !~ m/^\s*$ignore_priv\s*$/;

    # Update or delete .htaccess file according to $all
    my $htbase = "$EWOK_HOME/lib/htaccess";
    if (-f "$htbase.$label") {
        $htbase .= ".$label";
    }
    if ($all) {
        if (-f "$dirname/.htaccess") {
            return undef if ! unlink("$dirname/.htaccess");
        }
    }
    # Not viewable by ALL - set up .htaccess
    elsif (! -f $htbase || ! open HTBASE, "<", "$htbase") {
        die "[authzSave] cannot create .htaccess for $dirname: opening $htbase failed: $!";
    }
    elsif (!  open HTACCESS, ">", "$dirname/.htaccess") {
        die "[authzSave] cannot create .htaccess for $dirname: opening $dirname/.htaccess for write failed: $!";
    }
    # Opens okay, create the .htaccess
    else {
        my $htaccess;
        {
          local $/ = undef;
          $htaccess = <HTBASE>;
          close HTBASE;
        }
        # NONE privilege is never relevant, and we've already checked ALL
        my @users = grep !/^NONE|ALL$/, keys %$authz;
        # Omit any users with only $ignore_priv
        @users = grep { $authz->{$_} ne $ignore_priv } @users;
        # Create 'require user' line
        $htaccess .= 'require user ' . join(' ', @users) . "\n";
        # Save
        print HTACCESS $htaccess;
        close HTACCESS;
    }

    return 1;
}

#
# Check all @EWOK_STAGE_DIRS get initialised with default permissions
#
BEGIN
{
    if (@EWOK_ADMINS) {
        for my $dir (@EWOK_STAGE_DIRS) {
            if (-d $dir && ! -f "$dir/$EWOK_AUTHZ_OBJECT") {
                my $default = {
                    ALL => 'viewpubl',
                    NONE => 'view',
                };
                # Default admins to permissions access (basically for view)
                $default->{$_} = 'permissions' for @EWOK_ADMINS;
                # Save
                authzSave(undef, $dir, $default);
            }
        }
    }
}

1;
