# tag: ewok file operations module

package Ewok::File;

require Exporter;
use File::Basename;
use File::Compare;
use File::Copy;
use File::Find;
use File::DirSync;
use HTML::EmbperlObject;
use URI::Escape;
# use POSIX qw(strftime);
use Carp;
use Ewok::Config qw(:all);
use Ewok::Util qw(:all);
use Ewok::Meta qw(:all);
use Ewok::Authz qw(authzObjCheck authzSave authzGetAll authzGetUserHash);
use Ewok::TextMarkup qw(markup);
use Ewok::Locker;
use Ewok::Logger;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
    &readdir
    &autoindex
    &relativise
    &upload
    &create
    &remove
    &rename
    &cutcopy
    &pastefiles
    &pastecheck
    &paste
    &targetRewrite
    &getPublishDirs
    &getPublishHash
    &publish
    &publishAll
    &restrictCheck
);
%EXPORT_TAGS = (
    all => [@EXPORT_OK],
    file => [qw(&upload &remove &rename &cutcopy 
        &pastecheck &paste &restrictCheck)],
    publish => [qw(&targetRewrite &getPublishDirs &getPublishHash 
        &publish &publishAll)],
);

use strict;

#
# Loads the publish map from $EWOK_PUBLISH_OBJECT and returns it.
#
sub loadPublishMap
{
    my $publish = fileLoad($EWOK_PUBLISH_OBJECT,"$EWOK_HOME/htdocs",
        { eval => 1, name => 'publish', cache => 1, notfound => {} });
    return undef if ref $publish ne 'HASH';
    return $publish;
}

#
# Saves $map back to $EWOK_PUBLISH_OBJECT.
# 
sub savePublishMap
{
    my $map = shift;
    croak "[savePublishMap] map argument not hashref" if ref $map ne 'HASH';
    return fileSave($map, $EWOK_PUBLISH_OBJECT, "$EWOK_HOME/htdocs", 
        { name => 'EWOK_PUBLISH_MAP' });
}

#
# Return an arrayref of directories and an arrayref of files comprising the
#   contents of the given directory.
#
sub readdir
{
    my ($fs_dir, $arg) = @_;
    my $web_dir = $arg->{web_dir};
    my $admin = $arg->{admin};
    my $type = $arg->{type};
    my $add_only = $arg->{add_only};
    $web_dir = $fs_dir if ! defined $web_dir;

    my (@f, @d);
    # Read files in (unless $add_only is set)
    if (! $add_only) {
        return ([], []) if ! -d $fs_dir;
        return ([], []) if ! opendir DIR, $fs_dir;
        my @files = CORE::readdir(DIR);
        closedir DIR;

        # Omit dot files if not admin
        @files = grep(!/^\./, @files) if ! $admin;
        @files = grep(!/^\.{1,2}$/, @files) if $admin;
        # Divide into directories and files
        for my $entry (sort @files) {
            if (-d "$fs_dir/$entry") {
                push @d, $entry if ! $type || $type ne 'F';
            } else {
                push @f, $entry if ! $type || $type ne 'D';
            }
        }
    }

    # Add or remove special-case directories
    if (! $type || $type ne 'F') {
        @d = grep { my $foo = $_; ! grep /^$foo$/, @EWOK_ALL_DIRS_HIDE } @d;
        if ($web_dir eq '/' || $web_dir eq '') {
            @d = grep { my $foo = $_; ! grep /^$foo$/, @EWOK_TOP_DIRS_HIDE } @d;
            push @d, 'ewok' if $admin;
            @d = sort @d, @EWOK_TOP_DIRS_ADD;
        }
    }

    return \@d, \@f;
}

#
# Create or delete autoindex links from $fs_dir, recursively if $recursive 
#   is true;
#
sub autoindex ($;$$$) {
    my $fs_dir = shift;
    my $errors = shift;
    my $recursive = shift;
    my $delete = shift;
    $errors = [] if ! ref $errors eq 'ARRAY';
    my $errorCount = scalar @$errors;
    return 1 if ! $Ewok::Config::EWOK_AUTOINDEX_PATH;

    my @dir = ( $fs_dir );
    while ( @dir ) {
        my $d = shift @dir;
        if ($recursive) {
            my ($next) = Ewok::File::readdir($d, { type => 'D' });
            push @dir, map { "$d/$_" } @$next if @$next;
        }
        # Create
        if (! $delete) {
            if (! -e "$d/$Ewok::Config::EWOK_AUTOINDEX_OBJECT") {
                if (! symlink($Ewok::Config::EWOK_AUTOINDEX_PATH, "$d/$Ewok::Config::EWOK_AUTOINDEX_OBJECT")) {
                    warning("Error on autoindex create ($d)", $!, $errors);
                }
            }
            elsif (! -l "$d/$Ewok::Config::EWOK_AUTOINDEX_OBJECT") {
                warning("Error on autoindex create ($d): $Ewok::Config::EWOK_AUTOINDEX_OBJECT already exists.", $errors);
            }
        }
        # Delete
        else {
            if (-l "$d/$Ewok::Config::EWOK_AUTOINDEX_OBJECT") {
                if (! unlink("$d/$Ewok::Config::EWOK_AUTOINDEX_OBJECT")) {
                    warning("Error on autoindex delete ($d).", $!, $errors);
                }
            }
            elsif (-e "$d/$Ewok::Config::EWOK_AUTOINDEX_OBJECT") {
                 warning("Error on autoindex delete ($d): $Ewok::Config::EWOK_AUTOINDEX_OBJECT is not a link.", $errors);
            }
        }
    }

    return (scalar(@$errors) > $errorCount) ? undef : 1;
}

#
# Returns an arrayref of siblings to $fs_file and a hashref containing
#   their titles. Does not return $fs_file itself unless the 'self' arg
#   is set in the $arg hash, in which case the title is always 'SELF' for 
#   easy identification.
#
sub getSiblings
{
    my ($req, $fs_file, %arg) = @_;
    my $sort = $arg{'sort'};
    my $self = $arg{self};
    my $fs_dir = $fs_file;
    $fs_dir =~ s|[^/]+/*$||;
    return undef if ! -d $fs_dir;

    my @sib = ();
    my %title = ();
    my $meta = {};
    my ($d, $f) = Ewok::File::readdir($fs_dir, { type => 'F' });
    for my $file (@$f) {
        # Check for self
        if ("$fs_dir$file" eq $fs_file) {
            next if ! $self;
            push @sib, $file;
            $title{$file} = 'SELF';
        }
        else {
            $meta->{$file} = getMeta("$fs_dir$file");
            if (ref $meta eq 'HASH' && $meta->{$file}->{title}) {
                push @sib, $file;
                $title{$file} = $meta->{$file}->{title};
            }
        }
    }

    @sib = sort { 
        my $cmp = ($meta->{$a}->{item_order} || 5) cmp ($meta->{$b}->{item_order} || 5); 
        return $cmp if $cmp != 0; 
        $cmp = $meta->{$a}->{title} cmp $meta->{$b}->{title}; 
        return $cmp if $cmp != 0; 
        $a cmp $b 
    } @sib if $sort;

    return wantarray ? (\@sib, \%title) : \@sib;
}

#
# Return the absolute target $t made relative to directory $d
#   e.g. /a/b/c/d/e.html from /a/b/c/x/y/z is ../../../d/e.html
#
sub relativise ($$)
{
    my ($t, $d) = @_;
    # Return as is if $t is not absolute
    return $t if $t !~ m|^\s*/|;

    my $t1 = $t;
    $t1 =~ s|/[^/]*$|| if ! -d $t;
    my $d1 = $d;
    $d1 = substr($d1,0,-1) if substr($d1,-1) eq '/';
    my $count = 0;
    while ($d1) {
        # If d1 now matches the beginning of $t1, we're done
        last if $t1 =~ m/^$d1/;
        # Remove the final chunk
        $d1 =~ s|/[^/]+$||;
        $count++;
    }
    return undef if ! $d1;

    # Remove the matching prefix from $t
    $t =~ s|^$d1/?||;
    $t = '../' x $count . $t;
    $t = substr($t,0,-1) if substr($t,-1) eq '/';
    # Special case for current directory
    $t = '.' if $t eq '';

    return $t;
}

#
# Check for restrictions on special objects - either in %EWOK_RESTRICTED_OBJECTS
#   or metadata (=~ /\.meta$/). Called by create, upload, edit, remove, 
#   rename, cutcopy, paste, and publish.
# Returns true if okay to go ahead, false otherwise, with errors placed
#   in $errors arrayref, if supplied.
#
sub restrictCheck {
    my ($req, $op, $fs_dir, $filelist, $arg) = @_;
    my $errors = $arg->{errors} || [];
    $errors = [] if ref $errors ne 'ARRAY';
    my $authz = $arg->{authz};
    my $web_dir = $arg->{web_dir};
    my $in_wd = $web_dir ? " in " . substr($web_dir,0,-1) : '';
    my $prune = $arg->{prune};
    $op ||= 'manipulate';

    # Check general privilege levels
    $authz ||= authzGetUserHash($req, $fs_dir);
    if (ref $authz ne 'HASH' || 
           ($fs_dir ne $EWOK_CLIPBOARD && ! $authz->{edit}) || 
           ($op eq 'publish' && ! $authz->{publish})) {
        info("Error: insufficient privilege to $op$in_wd.", 
            "(files: " . join(',',@$filelist) . ')', $errors);
        return 0;
    }

    # Check the $fs_dir directory is valid
    my $valid = 0;
    for my $dir (@EWOK_STAGE_DIRS_RX, @EWOK_PUBL_DIRS_RX, $EWOK_CLIPBOARD) {
        $valid = 1, last if $fs_dir =~ m/$dir/;
    }
    if (! $valid) {
        my $dir = " '$web_dir'" if $web_dir;
        info("Error: cannot $op - directory$dir is not an ewok directory.",
          $errors);
        return 0;
    }

    my $okay = 1;
    my @newlist = ();
    for my $path (@$filelist) {
        # Paranoia
        $path = uri_unescape($path);
        my $file = basename($path);
        my $meta = qr/\.meta$/;
        # Fail by default on restricted objects - add specific allow cases first
        if ($EWOK_RESTRICTED_OBJECTS{$file} || 
                $file =~ m/$meta/ || 
                $file =~ m/$EWOK_TEMPLATE_RE/) {
            # Add $path to @newlist
            push @newlist, $path;
            # Anyone can create/delete/publish autoindex files
            next if $file eq $EWOK_AUTOINDEX_OBJECT && 
                ($op eq 'create' || $op eq 'delete' || $op eq 'publish');
            # Anyone can delete or trash metadata
            next if $file =~ /$meta/ && ($op eq 'delete' || $op eq 'trash');
            # People with 'permissions' priv can delete/cut/copy/paste authz.map
            next if ($file eq $EWOK_AUTHZ_OBJECT || $file eq '.htaccess') && 
                 $authz->{permissions} &&
                ($op eq 'delete' || $op eq 'cut' || $op eq 'trash' || 
                 $op eq 'copy' || $op eq 'paste');
            # People with 'templates' priv can do anything to templates except publish
            next if $file =~ m/$EWOK_TEMPLATE_RE/ && 
                 $authz->{templates} && $op ne 'publish';

            # Admins can only edit $EWOK_PUBLISH_OBJECT
            next if $file eq $EWOK_PUBLISH_OBJECT && $authz->{admin} &&
                $op eq 'edit';
            # Admins can do most things to %EWOK_ADMIN_OBJECTS
            next if $EWOK_ADMIN_OBJECTS{$file} && $authz->{admin} &&
                ($op eq 'create' || $op eq 'upload' || $op eq 'edit' ||
                 $op eq 'delete' || $op eq 'cut' || $op eq 'trash' || 
                 $op eq 'copy' || $op eq 'paste');
            # Admins can cut/copy/paste/rename metadata
            next if $file =~ /$meta/ && $authz->{admin} &&
                ($op eq 'cut' || $op eq 'copy' || $op eq 'paste' || 
                 $op eq 'paste as' || $op eq 'rename');

            # Fail - note error or (quietly) prune
            $okay = 0;
            if ($prune) {
                pop @newlist;
            }
            else {
                info("Error: cannot $op restricted file $file.", $errors);
            }
        }
        else {
            push @newlist, $path;
        }
    }

    # If prune replace filelist with newlist, if different
    if ($prune) {
        $filelist = splice @$filelist, 0, $#$filelist+1, @newlist if ! $okay;
        return 1;
    }
    # Otherwise return status
    else {
        return $okay;
    }
}

#
# Actions performed after the creation of a new object (currently 
#   auto_publishing). Called by create, renameOne, upload, and paste.
#
sub postNew ($$$;$) {
    my ($req, $fs_dir, $filelist, $arg) = @_;
    $filelist = [ $filelist ] if ! ref $filelist;
    my $dirmeta = $arg->{dirmeta};
    my $web_dir = $arg->{web_dir} || $fs_dir;
    my $errors = $arg->{errors} || [];

    $dirmeta = getMeta($fs_dir, { type => 'dir', inherit => 'I'}) 
        if ! ref $dirmeta;
    return if ! $dirmeta->{auto_publish};

    # Split suffixes and remove any leading '.'
    my @list = map { (substr($_, 0, 1) eq '.') ? substr($_,1) : $_ } split /\s*[,|\s]\s*/, $dirmeta->{auto_publish};
    my $autopublish = join('|',@list);

    # Publish
    $filelist = [ $filelist ] if ref $filelist ne 'ARRAY';
    publish($req, $filelist, $fs_dir, 
        { regex => ".$autopublish\$", errors => $errors });
}

#
# Actions performed after the deletion of an object (currently 
#   auto_publishing).
# Called by delete and cutcopy.
#
sub postDelete ($$$;$) {
    my ($req, $fs_dir, $filelist, $arg) = @_;
    my $dirmeta = $arg->{dirmeta};
    my $web_dir = $arg->{web_dir} || $fs_dir;
    my $errors = $arg->{errors} || [];

    $dirmeta = getMeta($fs_dir, { type => 'dir', inherit => 'I' }) 
        if ! ref $dirmeta;
    return if ! $dirmeta->{auto_publish};

    # Split suffixes and remove any leading '.'
    my @list = map { (substr($_, 0, 1) eq '.') ? substr($_,1) : $_ } split /\s*[,|\s]\s*/, $dirmeta->{auto_publish};
    my $autopublish = join('|',@list);

    # Unpublish 
    unpublish($req, $fs_dir, $filelist, 
            { regex => "\.$autopublish\$", errors => $errors });
}

#
# Upload $infile to $fs_dir
#
sub upload ($$$;$) 
{
    my ($req, $fs_dir, $infile, $arg) = @_;
    my $web_dir = $arg->{web_dir} || $fs_dir;
    my $errors = $arg->{errors} || [];
    my $errorCount = scalar @$errors;

    my $name = $infile;
    $name =~ s|^.*[/\\]||;
#   elog($req,"[upload] $infile -> $fs_dir ($name)");

    # Check for restricted objects
    return undef 
        if ! restrictCheck($req, 'upload', $fs_dir, [ $name ], { errors => $errors });

    # Need to get a lock on the destination file
    my $l = Ewok::Locker->new();
    if ($l->lock("$fs_dir$name", $req->{user}, { errors => $errors,
            altfile => "$web_dir$name", op => 'upload' })) {     
        if (open INFILE, ">", "$fs_dir$name") {
            my $buffer;
            print INFILE $buffer
                while read($infile, $buffer, 32768);
            close INFILE;
            if ($name !~ m/$Ewok::Config::EWOK_NONMETA_RE/i) {
                # Setup default metadata
#               my $meta = getMetaCreate("$fs_dir$name");
                setMeta(undef, "$fs_dir$name", { create => 1 });
            }
            # Call postNew
#           postNew($req, $fs_dir, $name);
        }
        else {
            warning("Error: file upload failed.", $!, $errors);
        }
        $l->unlock("$fs_dir$name", $req->{user},
            { altfile => "$web_dir$name" });
    } 
    return (scalar(@$errors) > $errorCount) ? undef : 1;
}

#
# Return a (hopefully unique) filename using a system-wide id counter.
#   $style indicates the naming style to be used - vanilla id, or modified
#   with the year or year/month. A sprintf template may also be given to 
#   somewhat customise the name generated.
# Returns undef on failure, with errors in $errors arrayref, if given.
#
# TODO: this is obsolete, reliant on database. Revisit.
#
sub generateName
{
    return;

    my ($req, $style, $arg) = @_;
    my $template = $arg->{template};
    my $dir = $arg->{dir};
    my $errors = $arg->{errors};
    $errors = [] if ref $errors ne 'ARRAY';

    $style = lc($style);
    $template ||= ($style eq 'id') ? 'file-%06d.html' : '%s-%06d.html';
    my @styles = qw(id year quarter month day);
    if (!grep /^$style$/, @styles) {
        info("Error: invalid name style '$style'.", $errors);
        return undef;
    }

    # Generate our id
#   my $set;
#   *set = DBIx::Recordset->Insert({
#       '!DataSource'   => 'ewok',
#       '!Table'        => 'id',
#       '!Serial'       => 'id',
#       dir             => $dir,
#   });
#   my $id = $set->LastSerial();
#   if (! $id) {
#       warning("Error: failure generating new file id.");
#       return undef;
#   }

    # Styles - generate extra tag
#   my $tag;
#   if ($style eq 'year') {
#       $tag = strftime("%Y", localtime);
#   }
#   elsif ($style eq 'quarter') {
#       $tag = strftime("%Yq", localtime);
#       $tag .= int((strftime("%m", localtime) - 1) / 3) + 1;
#   }
#   elsif ($style eq 'month') {
#       $tag = strftime("%Y%m", localtime);
#   }
#   elsif ($style eq 'day') {
#       $tag = strftime("%Y%m%d", localtime);
#   }

    # Apply the template
#   return ($style eq 'id') ? sprintf($template, $id) : 
#                             sprintf($template, $tag, $id);
}

#
# Create an element of $type ([dfl] for dir/file/link) called $name in $fs_dir.
# Return the name of the element on success (esp. useful if generated), or 
#   undef on failure, with error messages in the $errors arrayref if given.
#
sub create
{
    my $arg = shift;
    carp "[create] argument not hashref" if ref $arg ne 'HASH';
    my $req = $arg->{req} || {};
    my $fs_dir = $arg->{fs_dir} || $arg->{dir};
    my $name_in = $arg->{name};
    my $name_style = $arg->{name_style};
    my $name_template = $arg->{name_template};
    my $type = $arg->{type} || 'f';
    my $errors = $arg->{errors};
    my $wd = $arg->{wd} || $fs_dir;
    my $target = $arg->{target};
    my $text = $arg->{text};
    my $meta = $arg->{meta};
    my $mode = $arg->{mode};
    my $markup_text = $arg->{markup_text};
    my $text_filter = $arg->{text_filter};
    my $text_filter_args = $arg->{text_filter_args};

    $fs_dir = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    # Escape, trim and untaint $name
    my ($name) = (uri_unescape($name_in) =~ m/\s*(.*?)\s*$/);

    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find directory $wd.", $errors)
        if (! -d $fs_dir);
    info("Error: cannot write to directory $wd.", $errors)
        if (! @$errors && ! -w $fs_dir);
    info("Error: unknown type '$type' to create.", $errors)
        if ($type !~ m/^[dlf]$/);
    info("Error: element name cannot contain directory components.", $errors)
        if ($name =~ m|/|);
    info("Error: no name argument given to create.", $errors)
        if (! $name && ! $name_style);
    info("Error: cannot set both 'name' and 'name_style' arguments.", $errors)
        if ($name && $name_style);
    info("Error: cannot create hidden elements.", $errors)
        if ($name =~ m/^\./);
    info("Error: no target given for link.", $errors)
        if ($type eq 'l' && ! $target);
    info("Error: cannot give text when creating " . 
            (($type eq 'l') ? 'link' : 'directory') . '.', $errors)
        if ($text && ($type eq 'l' || $type eq 'd'));
    restrictCheck($req, 'create', $fs_dir, [ $name ], { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    my $dirmeta = getMeta($fs_dir, { type => 'dir', inherit => 'I' });
    $name = generateName($req, $name_style, { errors => $errors, 
        template => $name_template, dir => $fs_dir }) if $name_style;
    info("Error: $wd/$name already exists.", $errors)
        if ! @$errors && -e "$fs_dir/$name";
    return undef if scalar(@$errors) > $errorCount;

    # File create
    if ($type eq 'f') {
        my $fs_file = "$fs_dir/$name";
        if ($name eq $Ewok::Config::EWOK_AUTOINDEX_OBJECT) {
            return undef if ! autoindex($fs_dir, $errors);
        }
        elsif (! open FILE, ">>", $fs_file) {
            warning("Error: unable to create $wd/$name.", $errors);
            return undef;
        }
        else {
            if ($text) {
                $text = markup(input => $text) if $markup_text;
                print FILE $text;
            }
            close FILE;
            chmod $mode, $fs_file if $mode;
            if ($fs_file !~ m/$Ewok::Config::EWOK_NONMETA_RE/i) {
                # Setup default metadata
                setMeta($meta, $fs_file, { create => 1, errors => $errors });
            }
        }
    }

    # Symlink create
    elsif ($type eq 'l') {
        my $wt = $target;
        # Map (web-relative) absolute target to filesystem
        if ($target =~ m|^\s*/|) {
            if (ref $req->{req_rec}) {
                $target = $req->{req_rec}->lookup_uri($target)->filename;
            }
            else {
                info("Error: unable to map $wt - please try a relative link.",
                  $errors);
                return undef;
            }
        }
        # Or map relative target to filesystem
        else {
            if ($wd ne $fs_dir && ref $req->{req_rec}) {
                $target = "$wd/$target";
                $target = $req->{req_rec}->lookup_uri($target)->filename;
            }
            # If we can't do a lookup try and fake it - this could be dangerous
            else {
                $target = "$fs_dir/$target";
            }
        }

        # Now turn the absolute fs target into a relative one
        $target = relativise($target, $fs_dir);

        # Check target permissions
        my $authz = authzGetUserHash($req, "$fs_dir/$target");
        if (! $authz->{view}) {
            info("Error: cannot create link to $wt - you do not have view permissions.", $errors);
            return undef;
        }
        # Check target exists
        elsif (! -e "$fs_dir/$target") {
            info("Error: cannot create link to $wt - does not exist.", $errors);
            return undef;
        }
        # Check target isn't a link
        elsif (-l "$fs_dir/$target") {
            info("Error: cannot create link to $wt - it's a link.", $errors);
            return undef;
        }
        # Make symlink
        elsif (! symlink($target, "$fs_dir/$name")) {
            warning("Error creating link to $wt.", $!, $errors);
            return undef;
        }
        elsif ("$fs_dir/$name" !~ m/$Ewok::Config::EWOK_NONMETA_RE/i) {
            # Setup default metadata
            setMeta($meta, "$fs_dir/$name", { create => 1, errors => $errors });
        }
    }
     
    # Directory create
    elsif ($type eq 'd') {
        if (! mkdir("$fs_dir/$name", 0750)) {
            warning("Error on directory create.", $!, $errors);
            return undef;
        }

        # Setup default metadata
        setMeta($meta, "$fs_dir/$name", { create => 1, errors => $errors });

        # Check for autoindexing
        if (-f $Ewok::Config::EWOK_AUTOINDEX_PATH && $dirmeta->{auto_index}) {
            return undef if ! autoindex("$fs_dir/$name", $errors);
        }
    }

#   postNew($req, $fs_dir, $name, { dirmeta => $dirmeta });

    return $name;
}

#
# Callback for directory unlinks via File::Find
#
sub remove_unlink
{
    my $obj = $File::Find::name;
    $obj ||= shift;
    ($obj) = ($obj =~ m|^\s*(.*?)\s*$|);
    if (-d $obj) {
        rmdir $obj;
    }
    elsif (-e $obj) {
        unlink $obj;
    }
    else {
        croak("[remove_unlink] '$obj' not found");
    }
}

#
# Remove the given elements in $fs_dir: the directories in @$dirlist, and the
#   files in @$filelist. Directory has to be emtpy to be deleted (although
#   we delete dotfiles automatically).
# Returns true on success or undef on failure, with errors returned in the
#   @$errors arrayref (if provided).
#
sub remove
{
    my $req = shift;
    my $fs_dir = shift;
    my $filelist = shift;
    my $arg = shift;
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $unlink = $arg->{unlink};
    $filelist = [] if ! $filelist;
    $filelist = [ $filelist ] if ref $filelist ne 'ARRAY';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    $web_dir ||= $fs_dir;
    $fs_dir  = substr($fs_dir, 0, -1)  if substr($fs_dir, -1) eq '/';
    $web_dir = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find directory '$web_dir'.", $errors)
        if (! -d $fs_dir);
    info("Error: cannot write to directory '$web_dir'.", $errors)
        if (! @$errors && ! -w $fs_dir);
    restrictCheck($req, 'delete', $fs_dir, $filelist, { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    # Move to Trash (on the clipboard) unless $unlink is set
    if (!$unlink) {
        my $filelist2 = [ @$filelist ];
        # Explicitly discard any $EWOK_AUTOINDEX_OBJECT (cut would fail)
        if (grep /^$EWOK_AUTOINDEX_OBJECT$/, @$filelist) {
            $filelist2 = [ grep(!/^$EWOK_AUTOINDEX_OBJECT$/, @$filelist) ];
            unlink "$fs_dir/$EWOK_AUTOINDEX_OBJECT";
        }
        if (@$filelist2) {
            # Cut $filelist2 to a Trash area
            cutcopy($req, $fs_dir, $filelist2, { 
                tag => "Trash-" . $req->{user}, 
                op => 'trash', 
                errors => $errors, 
                web_dir => $web_dir, 
                keep => 1
            });
        }
        # Do any postDelete stuff (original $filelist)
#       postDelete($req, $fs_dir, $filelist, { errors => $errors });
    }

    # Explicit unlink
    else {
        for my $name (@$filelist) {
            # Paranoia check
            if ($name =~ m|/|) {
                warning("Error: '$name' is not a basename ('/' is not allowed).", $errors)
            }
            else {
                # Untaint $name
                ($name) = ($name =~ m|([^/]+)$|);
                # For directories, do a recursive delete, no locking
                if (-d "$fs_dir/$name") {
                    $ENV{PATH} = '/usr/bin:/bin';
                    find({
                        wanted => \&remove_unlink, 
                        no_chdir => 1,
                        untaint => 1,
                        bydepth => 1,
                    }, "$fs_dir/$name");
                }
                else {
                    # For files, need to worry about locking
                    my $l = Ewok::Locker->new();
                    if ($l->lock("$fs_dir/$name", $req->{user},
                            { errors => $errors, altfile => "$web_dir/$name",
                              op => 'delete' })) {
                        if (! unlink "$fs_dir/$name") {
                            warning("Error: unable to delete file '$web_dir/$name'.", $errors);
                        }
                        else {
                            # Remove metadata
                            delMeta("$fs_dir/$name");
                        }
                        $l->unlock("$fs_dir/$name", $req->{user},
                            { altfile => "$web_dir/$name" });
                    }
                }
            }
        }
#       postDelete($req, $fs_dir, $filelist, { errors => $errors })
#           if ! @$errors;
    }

    return undef if @$errors;
    return 1;
}

#
# Rename (within $web_dir) the set of files in @$namesFrom to @$namesTo (uses
#   &renameOne).
# Returns true if all renames are successful, undef otherwise, with errors
#   in the @$errors arrayref (if provided).
#
sub rename ($$$$;$$)
{
    my $req = shift;
    my $namesFrom = shift;
    my $namesTo = shift;
    my $fs_dir = shift;
    my $errors = shift;
    my $web_dir = shift;
    $web_dir ||= $fs_dir;
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find directory '$web_dir'.", $errors)
        if (! -d $fs_dir);
    info("Error: cannot write to directory '$web_dir'.", $errors)
        if (! @$errors && ! -w $fs_dir);
    info("Error: no objects given to rename.", $errors)
        if $#$namesFrom < 0;
    info("Error: rename from and to lists of different sizes.", $errors)
        if $#$namesFrom != $#$namesTo;
    restrictCheck($req, 'rename', $fs_dir, [ @$namesFrom, @$namesTo ], 
        { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    my %namesFrom = map { $_ => 1 } @$namesFrom;
    my (@nf, @nt);
    my $status = 1;
    my $last = $#$namesFrom;
    for (my $i = 0; $i <= $last; $i++) {
        # Use temporary files for targets that are also to be renamed
        if ($namesFrom{$namesTo->[$i]}) {
            push @$namesTo, $namesTo->[$i];
            $namesTo->[$i] .= '.tmp.' . $$;
            push @$namesFrom, $namesTo->[$i];
        }
        if (! renameOne($req, $namesFrom->[$i], $namesTo->[$i], $fs_dir, 
                $errors, $web_dir)) {
            $status = undef;
        }
    }
    # Second pass
    for (my $i = $last+1; $i <= $#$namesFrom; $i++) {
        if (! renameOne($req, $namesFrom->[$i], $namesTo->[$i], $fs_dir, 
                $errors, $web_dir)) {
            $status = undef;
        }
    }

    # postNew & postDelete
#   postNew($req, $fs_dir, $namesTo);
#   postDelete($req, $fs_dir, $namesFrom);

    return $status;
}

#
# Rename a single file/directory in $fs_dir from $nameFrom to $nameTo.
#   Returns true on success, undef on failure, with errors in the @$errors
#   arrayref (if provided).
#
sub renameOne ($$$;$$$)
{
    my $req = shift;
    my $nameFrom = shift;
    my $nameTo = shift;
    my $fs_dir = shift;
    my $errors = shift;
    my $web_dir = shift;
    $web_dir ||= $fs_dir;
    $fs_dir  = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    $web_dir = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find object '$web_dir/$nameFrom'.", $errors)
        if (! -e "$fs_dir/$nameFrom");
    info("Error: source filename '$nameFrom' contains invalid directory separator '/'.", $errors) if (uri_unescape($nameTo,'/') =~ m|/|);
    info("Error: target filename '$nameTo' contains invalid directory separator '/'.", $errors) if (uri_unescape($nameTo,'/') =~ m|/|);
    return undef if scalar(@$errors) > $errorCount;

    # Trim and untaint names
    $nameFrom =~ s|\s*(.*?)\s*$|$1|;
    ($nameFrom) = (uri_unescape($nameFrom) =~ m|.*?([^/]+)$|);
    $nameTo =~ s|\s*(.*?)\s*$|$1|;
    ($nameTo) = (uri_unescape($nameTo) =~ m|.*?([^/]+)$|);

    # Lock, rename, unlock
    my $l = Ewok::Locker->new();
    if ($l->lock("$fs_dir/$nameFrom", $req->{user},
                    { errors => $errors, altfile => "$web_dir/$nameFrom",
                      op => 'rename' })) {
        if ($l->lock("$fs_dir/$nameTo", $req->{user},
                    { errors => $errors, altfile => "$web_dir/$nameTo",
                      op => 'rename'  })) {
            if (!CORE::rename("$fs_dir/$nameFrom","$fs_dir/$nameTo")) {
                warning("Error: renaming '$nameFrom' to '$nameTo' in $web_dir " .
                  "failed.", $!, $errors);
            }
            # Move metadata for non-directories: get, set, delete
            elsif (! -d "$fs_dir/$nameFrom") {
                mvMeta($fs_dir, $nameFrom, $nameTo);
#               my $meta = getMeta("$fs_dir/$nameFrom");
#               if ($meta) {
#                   setMeta($meta, "$fs_dir/$nameTo");
#                   delMeta("$fs_dir/$nameFrom");
#               }
            }
            $l->unlock("$fs_dir/$nameTo", $req->{user},
                    { altfile => "$web_dir/$nameTo" });
        }
        $l->unlock("$fs_dir/$nameFrom", $req->{user},
                { altfile => "$web_dir/$nameFrom" });
    }

    return ((@$errors) ? undef : 1);
}

#
# Copy or move the files in @$filelist (in $fs_dir) to the $EWOK_CLIPBOARD
#   directory. Removes any files previously on the clipboard for this user
#   (identified with $tag) i.e. only one set of files is stored for a user.
# Returns true if all copies succeeded, undef on errors, with errors in the
#   errors arrayref (if supplied).
# Filenames on the clipboard are munged simply by prefixing '$tag-'.
#
sub cutcopy
{
    my $req = shift;
    my $fs_dir = shift;
    my $filelist = shift;
    my $arg = shift;
    my $tag = $arg->{tag};
    my $op = $arg->{op};
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $keep = $arg->{keep};
    my $clip_dir = $Ewok::Config::EWOK_CLIPBOARD;
    $web_dir ||= $fs_dir;
    $op ||= 'copy';
    $fs_dir   = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    $web_dir  = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    $clip_dir = substr($clip_dir, 0, -1) if substr($clip_dir, -1) eq '/';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Check error conditions
    warning("Error: clipboard directory '$clip_dir' does not exist (or is not writeable).", $errors)
        if ! -d $clip_dir || ! -w $clip_dir;
    info("Error: unknown operation '$op' ('copy', 'cut', and 'trash' supported).", $errors)
        if $op ne 'copy' && $op ne 'cut' && $op ne 'trash';
    warning("Error: required tag not set for $op.", $errors)
        if scalar(@$errors <= $errorCount) && ! $tag;
    info("Error: no files selected to $op.", $errors)
        if scalar(@$errors <= $errorCount) && ! @$filelist;
    restrictCheck($req, $op, $fs_dir, $filelist, { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    # Remove any existing files (and metadata) on the clipboard for this tag
    chdir ($clip_dir);
    if (! $keep) {
        my @delete = <$tag-*>;
        remove($req, $clip_dir, \@delete, {
            errors => $errors, 
            web_dir => '/clip', 
            unlink => 1,
        });
        return undef if scalar(@$errors) > $errorCount;
    }

    # Copy/move each file
    for my $file (@$filelist) {
        # Trim and untaint $file
        $file =~ s|\s*(.*?)\s*$|$1|;
        ($file) = (uri_unescape($file) =~ m|.*?([^/]+)$|);
        if (-f "$fs_dir/$file" || -l "$fs_dir/$file") {
            # Copy
            my $l = Ewok::Locker->new();
            if ($op eq 'copy') {
                if (! copy("$fs_dir/$file", "$clip_dir/$tag-$file")) {
                    warning("Error: $op failed.", $!, $errors);
                }
                else {
                    my $meta = getMeta("$fs_dir/$file");
                    setMeta($meta, "$clip_dir/$tag-$file") if $meta;
                }
            }
            # Cut or trash
            elsif ($l->lock("$fs_dir/$file", $req->{user},
                    { errors => $errors, altfile => "$web_dir/$file",
                      op => 'cut' })) {
                if (! move("$fs_dir/$file", "$clip_dir/$tag-$file")) {
                    warning("Error: $op failed.", $!, $errors);
                }
                else {
                    my $meta = getMeta("$fs_dir/$file");
                    delMeta("$fs_dir/$file");
                    delMeta("$clip_dir/$tag-$file") if $keep;
                    setMeta($meta, "$clip_dir/$tag-$file") if $meta;
                }
                $l->unlock("$fs_dir/$file", $req->{user},
                    { altfile => "$web_dir/$file" });
            }
        }
        # For directories, metadata is automatic and locking is prohibitive
        elsif (-d "$fs_dir/$file") {
            # Copy
            if ($op eq 'copy') {
                my $ds = File::DirSync->new({ nocache => 1 });
                $ds->dirsync("$fs_dir/$file", "$clip_dir/$tag-$file");
            }
            # Cut or trash
            else {
                if (-d "$clip_dir/$tag-$file") {
                    remove($req, $clip_dir, "$tag-$file", {
                        errors => $errors,
                        web_dir => '/clip',
                        unlink => 1,
                    });
                }
                if (! move("$fs_dir/$file", "$clip_dir/$tag-$file")) {
                    warning("Error: $op failed.", $!, $errors);
                }
            }
        }
        else {
            info("Error: file '$web_dir/$file' does not exist.", $errors);
        }
    }
#   postDelete($req, $fs_dir, $filelist, { errors => $errors }) 
#       if $op ne 'copy';

    return ! scalar(@$errors) > $errorCount;
}

#
# Returns an arrayref of files available on the clipboard for this $tag (user),
#   or undef on error.
#
sub pastefiles ($;$)
{
    my $tag = shift;
    my $munge = shift;
    my $clip_dir = $Ewok::Config::EWOK_CLIPBOARD;
    $clip_dir .= '/' if substr($clip_dir, length($clip_dir)-1, 1) ne '/';

    # Return undef if any errors
    return undef if ! -d $clip_dir || ! -r $clip_dir;
    return undef if ! $tag;

    # Glob $tag files on the clipboard
    chdir($clip_dir);
    my @filelist = sort <$tag-*>;
    map { s/^$tag-// } @filelist if $munge;
    return \@filelist;
}

#
# Checks whether there are any files to paste for this $tag (user).
# Returns the file count (0 if none), or undef on error.
#
sub pastecheck ($)
{
    my $tag = shift;
    my $filelist = pastefiles($tag);
    return (defined $filelist) ? scalar @$filelist : undef;
}

#
# Paste (copy) all $tag files to $fs_dir from the $EWOK_CLIPBOARD directory.
# Returns true if all copies are successful, undef otherwise, with errors
#   returned via the $errors arrayref (if provided).
# Filenames on the clipboard are munged simply by prefixing '$tag-'.
#
sub paste
{
    my $req = shift;
    my $fs_dir = shift;
    my $arg = shift;
    my $tag = $arg->{tag} || $req->{user};
    my $authz = $arg->{authz};
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $fromlist = $arg->{fromlist};
    my $tolist = $arg->{tolist};
    my $clobber = $arg->{clobber};
    my $clip_dir = $Ewok::Config::EWOK_CLIPBOARD;
    $web_dir ||= $fs_dir;
    $fs_dir   = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    $web_dir  = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    $clip_dir = substr($clip_dir, 0, -1) if substr($clip_dir, -1) eq '/';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;
    my %rename = ();

    # Check error conditions
    warning("Error: clipboard directory '$EWOK_CLIPBOARD' does not exist.", $errors)
        if ! -d $EWOK_CLIPBOARD || ! -r $EWOK_CLIPBOARD;
    info("Error: cannot write to paste directory '$web_dir'.", $errors)
        if ! -w $fs_dir;
    info("Error: required paste tag not set.", $errors)
        if ! $tag;
    return undef if scalar(@$errors) > $errorCount;

    # Glob $tag files on the clipboard
    # We treat no files to paste as an error (use &pastecheck first)
    chdir($clip_dir);
    my @filelist = @{pastefiles($tag)};
    info("Error: no files found to paste for '$tag'.", $errors)
        if ! @filelist;
    if (@$fromlist && @$tolist) {
        if (scalar @filelist != scalar @$fromlist ||
            scalar @filelist != scalar @$tolist) {
            warning("Error: found " . scalar(@filelist) . 
              "files to paste but rename lists contain " . scalar(@$fromlist) .
              " and " . scalar(@$tolist) . "respectively.", $errors);
        }
        else {
            for (my $i = 0; $i <= $#$fromlist; $i++) {
                $rename{"$tag-$fromlist->[$i]"} = $tolist->[$i];
            }
        }
    }
    return undef if ! authzObjCheck(\@filelist, $authz, 'paste', $errors);
    restrictCheck($req, 'paste', $fs_dir, \@filelist, { errors => $errors });
    restrictCheck($req, 'paste as', $fs_dir, $tolist, { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    # Paste each file
    for (my $i = 0; $i <= $#filelist; $i++) {
        # Untaint
        my ($filename) = ($filelist[$i] =~ m|([^/]+)$|);
        my $shortname = $filename;
        $shortname =~ s/^$tag-//;
        my $newname;
        if (!@$fromlist) {
            $newname = $shortname;
        }
        else {
            # Untaint
            ($newname) = (($rename{$filename} || $shortname) =~ m|([^/]+)$|);
        }
        if (-e "$clip_dir/$filename") {
            # If the target file exists, make a copy or overwrite
            if (-e "$fs_dir/$newname") {
                # Check whether to overwrite this file
                if (! $clobber->{$shortname}) {
                    my $newname2 = "Copy-Of-$newname";
                    if (-e "$fs_dir/$newname2") {
                        for my $i (2 .. 99) {
                            $newname2 = "Copy${i}-Of-$newname";
                            last if ! -e "$fs_dir/$newname2";
                        }
                        if (-e "$fs_dir/$newname2") {
                            warning("Error: paste failed (too many copies already).", $errors);
                            next;
                        }
                    }
                    $newname = $newname2;
                }
            }
            my $fs_file = "$fs_dir/$newname";
            # For directories, metadata is automatic, don't worry about locking
            if (-d "$clip_dir/$filename") {
                my $ds = File::DirSync->new({ nocache => 1 });
                $ds->dirsync("$clip_dir/$filename", $fs_file);
                # Call postNew for all files
#               $ENV{PATH} = '/usr/bin:/bin';
#               find({ 
#                   wanted => sub {
#                       postNew($req, $File::Find::dir, basename($_)) 
#                           if ! -d $File::Find::name;
#                   },
#                   no_chdir => 1,
#                   untaint => 1,
#                   bydepth => 1,
#               }, $fs_file);
            }
            else {
            # Lock the file we're pasting to
                my $l = Ewok::Locker->new();
                if ($l->lock($fs_file, $req->{user},
                        { errors => $errors, altfile => "$web_dir/$newname",
                          op => 'paste' })) {
                    if (! &copy("$clip_dir/$filename", "$fs_dir/$newname")) {
                        warning("Error: paste failed.", $!, $errors);
                    }
                    else {
                        if ($fs_file !~ m/$Ewok::Config::EWOK_NONMETA_RE/i) {
                            # Copy metadata
                            my $meta = getMeta("$clip_dir/$filename");
                            if (defined $meta) {
                                setMeta($meta, $fs_file);
                            }
                        }
                        # Call #ostNew
#                       postNew($req, $fs_dir, $newname);
                    }

                    # Unlock the pasted file
                    $l->unlock("$fs_dir/$newname", $req->{user},
                        { altfile => "$web_dir/$newname" });
                }
            }
        }
        else {
            warning("Error: clipboard file '$newname' not found.", $errors);
        }
    }
    return ((scalar(@$errors) > $errorCount) ? undef : 1);
}

#
# Rewrite $web_file using %EWOK_PUBLISH_MAP, but checking it against the
#   'target' directory rather than the 'source' one. This is useful for 
#   viewing remote target directory content.
#
sub targetRewrite ($$)
{
    my ($req, $web_file) = @_;

    # Iterate through the EWOK_PUBLISH_MAP
    my $target;
    my $uri = '';
    my $EWOK_PUBLISH_MAP = loadPublishMap();
    OUTER: for my $key (keys %$EWOK_PUBLISH_MAP) {
        my $map = $EWOK_PUBLISH_MAP->{$key};
        # If there's only one entry for this key, check the target directly
        if (ref $map eq 'HASH') {
            $target = $map->{dest_web_dir};
            next if ! $target || $web_file !~ m/^$target/;
            # Match! 
            $uri = $map->{dest_url};
            last;
        }
        # If multiple entries, check them all
        else {
            for my $m (@$map) {
                $target = $m->{dest_web_dir};
                next if ! $target || $web_file !~ m/^$target/;
                # Match! 
                $uri = $m->{dest_url};
                last OUTER;
            }
        }
    }

    # If found and $uri is defined, do the rewrite
    if ($uri) {
        # Check trailing slashes on $target and $uri
        if (substr($target,-1) eq '/' && substr($uri,-1) ne '/') {
            $uri .= '/';
        }
        elsif (substr($target,-1) ne '/' && substr($uri,-1) eq '/') {
            $uri = substr($uri,0,-1);
        }
        $web_file =~ s/^$target/$uri/;
    }

    return $web_file;
}

#
# Map current directory settings to publish directories
#
sub getPublishDirs 
{
    my $pubHash = shift;
    my $web_dir = shift;
    $web_dir .= '/' if substr($web_dir,-1) ne '/';

    return (undef, undef) if ref $pubHash ne 'HASH';

    my $src_web_dir = $pubHash->{src_web_dir};
    my $dest_web_dir = $pubHash->{dest_web_dir};
    my $dest_url = $pubHash->{dest_url};
    my $pub_dir = $web_dir;
    $pub_dir =~ s/^$src_web_dir/$dest_web_dir/;
    my $pub_url = $pub_dir;
    if ($dest_url) {
        ($pub_url = $web_dir) =~ s|^$src_web_dir|$dest_url/|;
    }

    # Clean up '/'s in $pub_url
    $pub_url =~ s|//+|/|g;
    $pub_url =~ s|:/|://|g;
    $pub_url =~ s|/$||;

    return ($pub_url, $pub_dir);
}

#
# Do a lookup on $fs_file in %$EWOK_PUBLISH_MAP. In scalar context returns 
#   the appropriate entry as a hashref with the keys:
#     - src_fs_dir
#     - src_web_dir
#     - dest_fs_dir
#     - dest_web_dir
#     - dest_url
#     - label
#     - template
#     - template_dir
# Uses $label to distinguish between multiple map entries for the same key, 
#   where necessary. 
# In array context returns an array containing hashrefs (with the same keys
#   as above) for all map entries for $fs_file.
#
sub getPublishHash ($;$) 
{
    my $fs_file = shift;
    my $label = shift;
    $label ||= 'Default' if ! wantarray;

    # Iterate in reverse key length order through %$EWOK_PUBLISH_MAP
    # This should maybe be reverse src_web_dir order, but that's much harder!
    my ($found, $a, $b);
    my $EWOK_PUBLISH_MAP = loadPublishMap();
    OUTER: 
    for my $key (sort {length $Ewok::File::b <=> length $Ewok::File::a} keys(%$EWOK_PUBLISH_MAP)) {
        next if $fs_file !~ m/^$key/;

        # Match!
        my $map = $EWOK_PUBLISH_MAP->{$key};
        # If there's only one entry for this key, we're done
        if (ref $map eq 'HASH') {
            # Found if $label not set or it matches
            if (! $label || $label eq 'Default' || uc($label) eq uc($map->{label})) {
                $found = $key;
                $a = $b = 0;
                last;
            }
            else {
                return undef;
            }
        }
        # Else look for the $label entry
        elsif ($label) {
            for my $i (0 .. $#$map) {
                if ($label eq 'Default' || uc($label) eq uc($map->[$i]->{label})) {
                    $found = $key;
                    $a = $b = $i;
                    last OUTER;
                }
            }
            # Fail if $label not found
            return undef;
        }
        # Else return all entries if wantarray
        elsif (wantarray) {
            $found = $key;
            $a = 0;
            $b = $#$map;
            last;
        }
    }

    return undef if ! $found;

    my @hashset = ();
    my $map = $EWOK_PUBLISH_MAP->{$found};
    $found .= '/' if substr($found,-1) ne '/';
    for my $i ($a .. $b) {
        my $hash = (ref $map eq 'ARRAY') ? $map->[$i] : $map;
        $hash->{src_fs_dir}   = $found;

        # Make sure all dir items finish with a '/'
        $hash->{src_web_dir} .= '/'  
            if $hash->{src_web_dir} && substr($hash->{src_web_dir},-1) ne '/';
        $hash->{dest_fs_dir} .= '/'  
            if $hash->{dest_fs_dir} && substr($hash->{dest_fs_dir},-1) ne '/';
        $hash->{dest_web_dir} .= '/' 
            if $hash->{dest_web_dir} && substr($hash->{dest_web_dir},-1) ne '/';
        $hash->{dest_url} .= '/'
            if $hash->{dest_url} && substr($hash->{dest_url},-1) ne '/';

        return $hash if ! wantarray;
        push @hashset, $hash;
    }

    return @hashset;
}

#
# Delete the files in @$filelist from the publish directory for $fs_dir.
#
sub unpublish ($$$;$) 
{
    my ($req, $fs_dir, $filelist, $arg) = @_;
    my $regex = $arg->{regex};
    my $web_dir = $arg->{web_dir};
    my $pubHash = $arg->{pubhash};
    my $errors = $arg->{errors} || [];
    my $errorCount = scalar @$errors;

    # Determine pubHash if not passed in
    $pubHash = &getPublishHash($fs_dir) if ref $pubHash ne 'HASH';

    my ($src_fs_dir, $dest_fs_dir);
    if (ref $pubHash eq 'HASH') {
        $src_fs_dir  = $pubHash->{src_fs_dir};
        $dest_fs_dir = $pubHash->{dest_fs_dir};
    }

    # Check for required parameters
    warning("Error: no files given to unpublish.", $errors) if ! @$filelist;
    warning("Error: publish source directory not set.", $errors) if ! $src_fs_dir;
    warning("Error: publish destination directory not set.", $errors) if ! $dest_fs_dir;
    return undef if scalar(@$errors) > $errorCount;

    # Check the top-level target directory exists
    my $fs_pub_dir = "$fs_dir/";
    $fs_pub_dir =~ s/^$src_fs_dir/$dest_fs_dir/;
    $fs_pub_dir =~ s|/+$||;
    if (! -d $fs_pub_dir) {
        warning("Error: publish directory $fs_pub_dir not found.", $errors);
        return undef;
    }

    # Unlink the given files if $fs_dir/$f does not exist
    for my $f (@$filelist) {
        next if $regex && $f !~ m/$regex/;
        next if -e "$fs_dir/$f";
        my $f = $fs_pub_dir . $f;
        if (-f $f || -l $f) {
            unlink $f;
        }
        # TODO: directories here!
    }

    # Do post_publish, if set
    if ($EWOK_ALLOW_EXEC && $pubHash->{post_publish}) {
        my $out = `$pubHash->{post_publish}`;
        info("post_publish: $out");
    }

    return 1;
}

#
# Delete any files that exist in $fs_pub_dir that do not exist in $fs_dir.
#
sub syncDeletes ($$$;$)
{
    my ($req, $fs_dir, $fs_pub_dir, $arg) = @_;
    my $test = $arg->{test};
    my ($d1, $f1) = Ewok::File::readdir($fs_dir);
    my ($d2, $f2) = Ewok::File::readdir($fs_pub_dir);
    
    my @deleted = ();
    for my $f (@$f2) {
        ($f) = (uri_unescape($f) =~ m|.*?([^/]+)$|);
        if (! grep /^$f$/, @$f1) {
            push @deleted, $f;
            unlink "$fs_pub_dir/$f" if ! $test;
        }
    }

    return @deleted;
}

#
# For EMBPERL_FILESMATCH files, use EmbperlObject::Execute to publish. 
#   Used for non-links, for links if $copylinks set, and for 'auto.html'
#
sub publish_embperl
{
    my ($fs_dir, $f, $fs_pub_dir, $f2, $arg) = @_;
    my $force = $arg->{force};
    my $test = $arg->{test};
    my $req = $arg->{req};
    my $results = $arg->{results};
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $target = $arg->{target};
    my $template = $arg->{template};
    my $dest_host = $arg->{dest_host};
    my $src_fs_dir = $arg->{src_fs_dir};
    my $dest_fs_dir = $arg->{dest_fs_dir};
    my $template_dir = $arg->{template_dir};
    $template_dir .= ';' if $template_dir;
    my $debug = $EWOK_ERROR_LEVEL eq 'debug';

    my ($old, $new);
    $ENV{EWOK_PUBLISHED} = 1;
    $ENV{SCRIPT_NAME} = "$web_dir/$f";
    my $fdat = { t => $template } if $template;
    my $options = $ENV{EMBPERL_OPTIONS};
    my $addpath = "$template_dir$EWOK_HOME/htdocs/templates;$EWOK_HOME/htdocs";
    $addpath =~ s!//+!/!g;
    debug("[publish_embperl] execute: inputfile '$fs_dir/$f', template '$template', options '$options', object_base '$ENV{EMBPERL_OBJECT_BASE}', object_stopdir '$src_fs_dir', object_addpath '$addpath'") if $debug;
    HTML::EmbperlObject::Execute({
        inputfile => "$fs_dir/$f",
        output => \$new,
        fdat => $fdat,
        options => $options,
        object_base => $ENV{EMBPERL_OBJECT_BASE},
        object_stopdir => $src_fs_dir,
        object_addpath => $addpath,
    });
    # Trim leading and trailing spaces
    $new =~ s/^\s+//s;
    $new =~ s/\s+$/\n/s;
    # If no output, push error and return
    if ($new eq '' || $new eq "\n") {
        warning("Error: publishing file '$f' produced no output - not published.", $errors);
        return undef;
    }
    my $newsize1 = length($new);

    # Do URL rewrites if $EWOK_PUBLISH_REWRITES set
    if ($EWOK_PUBLISH_REWRITES && $target ne 'stage') {
        my $EWOK_PUBLISH_MAP = loadPublishMap();
        # Iterate in reverse key length order through %$EWOK_PUBLISH_MAP
        # This should maybe be reverse src_web_dir order?
        for my $key (reverse sort {length $a <=> length $b} keys %$EWOK_PUBLISH_MAP) {
            my $pubHash = getPublishHash($key, $target);
            next if ! defined $pubHash;
            my $from = $pubHash->{src_web_dir};
            my $to = $pubHash->{dest_url};
            if ($to) {
                # Remove host part if matches $dest_host
                $to =~ s|^$dest_host|/|;
            }
            else {
                $to = $pubHash->{dest_web_dir};
            }
            if (! $from || ! $to) {
                if (! $from) {
                    warning("Error: unable to derive sensible 'from' component from '$pubHash->{src_web_dir}' for URL rewriting.\n", $errors);
                }
                if (! $to) {
                    warning("Error: unable to derive sensible 'to' component from '$pubHash->{dest_web_dir}' for URL rewriting.\n", $errors);
                }
                return undef;
            }
            $new =~ s/(\b(href|src)\s*=\s*")$from/$1$to/ig;
            $new =~ s/(\burl\s*\(\s*)$from/$1$to/ig;
            my $host = $dest_host;
            $host =~ s|/+$||;
            $new =~ s|<!--\s*EWOK_REWRITE\s*-->\s*(.*)http://$ENV{HTTP_HOST}$from|$1$host$to|ig;
        }
    }
    if ($debug) {
        my $old = -s "$fs_dir/$f";
        debug("[publish_embperl] $fs_dir/$f publish sizes: old $old, new initial $newsize1, new url rewrites " . length($new));
    }

    # Check there are differences if previously published file exists
    if (! $force && -f "$fs_pub_dir/$f2" && open(OLD, "<", "$fs_pub_dir/$f2")) {
        {
          local $/ = undef;
          $old = <OLD>;
        }
        close OLD;
        # Trim leading and trailing spaces
        $old =~ s/^\s*//so;
        $old =~ s/\s*$/\n/so;
        # Check for differences
        if (! $old) {
            debug("[publish_embperl] old publish file $fs_pub_dir/$f2 apparently empty");
        }
        elsif ($new eq $old) {
            push @{$results->{warnings}}, "Warning: file '$f2' identical to current published version - not published.";
            debug("[publish_embperl] publish file $fs_pub_dir/$f2 unchanged - not publishing.");
            return 0;
        }
        else {
            debug("[publish_embperl] publish file $fs_pub_dir/$f2 differs from previous version - publishing");
        }
    }

    # Publish - do the update if not in test mode
    if (! $test) {
        debug("[publish_embperl] doing write to $fs_pub_dir/$f2");
        # Check whether $f2 contains any further subdirectories
        if ($f2 =~ m|/|) {
            my $subdir = $f2;
            $subdir =~ s|/[^/]+$||;
            if (! -d "$fs_pub_dir/$subdir") {
                if (! makedir("$fs_pub_dir/$subdir", { check_dir => $dest_fs_dir, errors => $errors })) {
                    debug("[publish_embperl] makedir of '$fs_pub_dir/$subdir' failed.", @$errors);
                    return undef;
                }
            }
        }
        # If target is a link, remove it explicitly
        unlink "$fs_pub_dir/$f2" if -l "$fs_pub_dir/$f2";
        # Write $new out to a file
        if (! open NEW, ">", "$fs_pub_dir/$f2") {
            warning("Error: opening file '$f2' for writing failed - not published.", $!, $errors);
            return undef;
        }
        elsif (! print NEW $new) {
            warning("Error: writing to file '$f2' failed.", $!, $errors);
            close NEW;
            return undef;
        }
        close NEW;
    }
    else {
        debug("[publish_embperl] test mode - skipping write of $fs_pub_dir/$f2");
    }

    return 1;
}

sub publish_link
{
    my ($fs_dir, $f, $fs_pub_dir, $f2, $arg) = @_;
    my $force = $arg->{force};
    my $test = $arg->{test};
    my $results = $arg->{results};
    my $errors = $arg->{errors};

    # For links, recreate in publish area (assumes *always* relative)
    my ($target) = (readlink("$fs_dir/$f") =~ m|(\S.*?)\s*$|);   #/
    if (! $force && -l "$fs_pub_dir/$f2" &&
            readlink("$fs_pub_dir/$f2") eq $target) {
        push @{$results->{warnings}}, "Warning: link '$f' identical to current published version - not published.";
        return 0;
    }
    elsif (! $test) {
        if (-f "$fs_pub_dir/$f2" || -l "$fs_pub_dir/$f2") {
            unlink "$fs_pub_dir/$f2"; 
        }
        elsif (-d "$fs_pub_dir/$f2") {
            info("Error: publish of link '$f' failed - '$f2' already exists as directory.", $errors);
            return undef;
        }
        if (! chdir $fs_pub_dir) {
            info("Error: chdir to publish directory failed.", $!, $errors);
            return undef;
        }
        if (!symlink $target, "$fs_pub_dir/$f2") {
            info("Error: publish of link '$f' failed.", $!, $errors);
            return undef;
        }
    }
    return 1;
}

#
# For non-EMBPERL_FILESMATCH files, simply publish as copy
#
sub publish_copy
{
    my ($fs_dir, $f, $fs_pub_dir, $f2, $arg) = @_;
    my $force = $arg->{force};
    my $test = $arg->{test};
    my $results = $arg->{results};
    my $errors = $arg->{errors};
    my $dest_fs_dir = $arg->{dest_fs_dir};

    if (! $force && -f "$fs_pub_dir/$f2" && 
        compare("$fs_dir/$f","$fs_pub_dir/$f2") == 0) {
        push @{$results->{warnings}}, "Warning: file '$f' identical to current published version - not published." if $f !~ m/.htaccess$/;
        return 0;
    }

    # Copy to $fs_pub_dir
    elsif (! $test) {
        if ($f2 =~ m|/|) {
            my $subdir = $f2;
            $subdir =~ s|/[^/]+$||;
            if (! -d "$fs_pub_dir/$subdir") {
                if (! makedir("$fs_pub_dir/$subdir", { check_dir => $dest_fs_dir, errors => $errors })) {
                    return undef;
                }
            }
        }
        # If target is a link, remove it explicitly
        unlink "$fs_pub_dir/$f2" if -l "$fs_pub_dir/$f2";
        if (! copy("$fs_dir/$f", "$fs_pub_dir/$f2")) {
            info("Error: publish (copy) of file '$f' failed.", $!, $errors);
            return undef;
        }
    }

    return 1;
}

#
# Publish the given $filelist from staging (the $src_fs_dir tree) to
#   production ($dest_fs_dir tree).
# Return undef if errors, true otherwise. Results are returned in the 
#   $results hashref, typically 'pubList', 'deletes', 'errors', and
#   'warnings'.
#
sub publish
{
    my $req = shift;
    my $filelist = shift;
    my $fs_dir = shift;
    my $results = shift;
    my %arg = @_;
    my $web_dir = $arg{web_dir};
    my $pubHash = $arg{pubhash};
    my $refresh = $arg{refresh};
    my $target = $arg{target};
    my $template = $arg{template};
    my $pubfilename = $arg{pubfilename};
    my $stage_replace = $arg{stage_replace};
    my $regex = $arg{regex};
    my $test = $arg{test};
    my $force = $arg{force};
    my $label = $arg{label};
    my $fdat = $arg{fdat};
    my $copylinks = $arg{copylinks};
    my $sync_deletes = $arg{sync_deletes};
    my $scope = $arg{scope} || 'selected';
    my $debug = $arg{debug};

    $filelist = [ $filelist ] if ref $filelist ne 'ARRAY';
    $web_dir = $fs_dir if ! defined $web_dir;

    # Remove any trailing slashes
    $fs_dir  =~ s|/+$||;
    $web_dir =~ s|/+$||;

    # Untaint
    ($target) = ($target =~ m/(\S.*?)\s*$/);

    # Errors, warnings, and pubList can be ignored if not passed as arrayrefs
    $results->{pubList} = [] if ref $results->{pubList} ne 'ARRAY';
    $results->{deleteList} = [] if ref $results->{deleteList} ne 'ARRAY';
#   $results->{errors} = [] if ref $results->{errors} ne 'ARRAY';
    $results->{warnings} = [] if ref $results->{warnings} ne 'ARRAY';
    my $errors = $results->{errors};
    my $errorCount = scalar @$errors;

    # Determine pubHash if not passed in
    $pubHash = &getPublishHash($fs_dir, $target) if ref $pubHash ne 'HASH';

    # Setup publish directories
    my ($fs_pub_dir, $src_fs_dir, $dest_fs_dir, $dest_host, $template_dir);
    if (ref $pubHash eq 'HASH') {
        $src_fs_dir  = $pubHash->{src_fs_dir};
        $template ||= $pubHash->{template};
        $template_dir = $pubHash->{template_dir};
        $template_dir ||= "$src_fs_dir/templates";
        # If publishing locally, fake it
        if ($target eq 'stage') {
            $dest_fs_dir = $src_fs_dir;
        }
        else {
            $dest_fs_dir = $pubHash->{dest_fs_dir};
            ($dest_host = $pubHash->{dest_url}) =~ s|^(https?://[^/]+/).*$|$1|;
        }
    }

    # Check for required parameters
    info("Error: no files given to publish.", $errors)
        if ! @$filelist && ! $sync_deletes;
    info("Error: publish source directory not set.", $errors)
        if ! $src_fs_dir;
    info("Error: publish destination directory not set.", $errors)
        if ! $dest_fs_dir;
    # Silently prune illegal files for 'dir', 'tree', and 'refresh' publishes
    my @prune = ( prune => 1 ) if $scope ne 'selected' || $refresh;
    restrictCheck($req, 'publish', $src_fs_dir, $filelist, { errors => $errors, @prune, });
    return undef if scalar(@$errors) > $errorCount;

    # Check we are publishing from within the $src_fs_dir tree
    warning("Error: directory '$web_dir' not within publish source tree '$src_fs_dir'.", $errors)
        if ("$fs_dir/" !~ m/^$src_fs_dir/);
    return undef if scalar(@$errors) > $errorCount;

    # Check the top-level target directory exists
    $fs_pub_dir = "$fs_dir/";
    $fs_pub_dir =~ s/^$src_fs_dir/$dest_fs_dir/ if $target ne 'stage';
    $fs_pub_dir =~ s|/+$||;
    return undef if ! makedir($fs_pub_dir, { check_dir => $dest_fs_dir, 
        errors => $errors, copylinks => $copylinks });

    # Setup logger
#   my $t = Ewok::Logger->new('File::publish');
#   $t->info("[publish] $req->{user}: $web_dir($fs_dir), " . join(':',@$filelist) . ", $sync_deletes");

    if (@$filelist) {
        # Checkpoint globals that can get clobbered on publish
        executeCheckpoint($req, $fdat);

        # Publish the individual files
        for my $f (@$filelist) {
            # Untaint
            ($f) = ($f =~ m|^\s*(.+)\s*$|); 
            # Check for print version
            my $meta = getMeta("$fs_dir/$f");
            my $fp;
            if ($meta->{print_version}) {
                ($fp = $f) =~ s/(\.[^\.]*)$/.print$1/;
            }

            # Setup of publish filename
            my $f2 = $pubfilename;
            if ($f2) {
                $f2 .= $f if -d "$fs_pub_dir/$f2";
            }
            else {
                $f2 = $f;
            }
            return undef if $regex && $f !~ m/$regex/;
            if (! -e "$fs_dir/$f") {
                info("Error: '$f' does not exist - not published.", $errors)
            }
            elsif (! -f "$fs_dir/$f" && ! -l "$fs_dir/$f") {
                info("Error: '$f' is not a file - not published.", $errors)
            }
            elsif ($f =~ m/\.epl$/i) {
                info("Warning: cannot publish '.epl' files - '$f' not published.", $results->{warnings});
            }

            next if scalar(@$errors) > $errorCount;

            if ($target eq 'stage' && -e "$fs_pub_dir/$f2" && ! $stage_replace) {
                info("Error: '$f2' already exists - select 'Force replace' option to overwrite.", $errors)
            }
            # Refresh only if existing
            elsif ($refresh && ! -e "$fs_pub_dir/$f2") {
                debug("refresh mode and '$fs_pub_dir/$f2' does not exist - not published");
                next;
            }

            # Do the actual publish
            my $publish = 0;
            if ($f =~ $ENV{EMBPERL_FILESMATCH} &&
              (! -l "$fs_dir/$f" || $copylinks || $f eq $EWOK_AUTOINDEX_OBJECT)) {
                $publish = publish_embperl($fs_dir, $f, $fs_pub_dir, $f2,
                    { test => $test, refresh => $refresh, 
                      req => $req, force => $force,
                      errors => $errors, results => $results, 
                      web_dir => $web_dir, target => $target, 
                      template => $template, dest_host => $dest_host, 
                      src_fs_dir => $src_fs_dir, dest_fs_dir => $dest_fs_dir,
                      template_dir => $template_dir, debug => $debug, });
                # Publish print version, if applicable
                if ($fp && $template ne 'print') {
                    my $print = publish_embperl($fs_dir, $f, $fs_pub_dir, $fp,
                        { test => $test, refresh => $refresh, 
                          req => $req, force => $force,
                          errors => $errors, results => $results, 
                          web_dir => $web_dir, target => $target, 
                          template => 'print', dest_host => $dest_host, 
                          src_fs_dir => $src_fs_dir, 
                          dest_fs_dir => $dest_fs_dir,
                          template_dir => $template_dir, debug => $debug, });
                    push @{$results->{pubList}}, $fp 
                        if $print && $f !~ m/.htaccess$/;
                }
            }
            elsif (-l "$fs_dir/$f" && !$copylinks) {
                $publish = publish_link($fs_dir, $f, $fs_pub_dir, $f2,
                    { test => $test, refresh => $refresh, 
                      force => $force, debug => $debug,
                      errors => $errors, results => $results, });
            }
            else {
                $publish = publish_copy($fs_dir, $f, $fs_pub_dir, $f2,
                    { test => $test, refresh => $refresh, 
                      force => $force, debug => $debug,
                      errors => $errors, results => $results, 
                      dest_fs_dir => $dest_fs_dir, });
            }

            # Success - add file to pubList
            push @{$results->{pubList}}, $f2 if $publish && $f !~ m/.htaccess$/;
        }

        # Restore those clobberable globals
        executeRestore($req, $fdat);
    }

    # Do syncDeletes
    push @{$results->{deleteList}}, syncDeletes($req, $fs_dir, $fs_pub_dir,
         { test => $test }) if $sync_deletes;

    # Do a 'publish' authzSave
    if ($Ewok::Config::EWOK_AUTHZ_PUBLISH && -f "$fs_dir/$EWOK_AUTHZ_OBJECT") {
        if (!authzSave($req, $fs_pub_dir, authzGetAll($req, $fs_dir), 
                { publish => $target, label => $label })) {
            warning("Error saving permissions object.", $!, $errors)
        }
    }

    # Do post_publish, if set
    if ($EWOK_ALLOW_EXEC && $pubHash->{post_publish} && ! $test) {
        my $out = `$pubHash->{post_publish}`;
        warning("post_publish: $out") if $out;
    }

    # Return true if no errors
    return ! scalar(@$errors) > $errorCount;
}

#
# Wrapper subroutine generating all the files to publish at or below the $fs_dir.
#   Calls &publish with this list.
#
sub publishAll
{
    my $req = shift;
    my $fs_dir = shift;
    my $results = shift;
    my %arg = @_;
    my $errors = $results->{errors};
    my $recursive = $arg{recursive};
    my $copylinks = $arg{copylinks};
    my $web_dir = $arg{web_dir};
    $arg{scope} = $recursive ? 'tree' : 'directory';

    $recursive = 0 if ! defined $recursive;
    $web_dir = $fs_dir if ! defined $web_dir;
    $errors = [] if ref $errors ne 'ARRAY';

    # Remove any trailing slashes
    $fs_dir  =~ s|/+$||;
    $web_dir =~ s|/+$||;
 
    # Check whether this directory is actually a link; if so, dereference it
    my $fs_link;
    if (-l $fs_dir && ! $copylinks) {
        $fs_link = $fs_dir;
        my $linkdir = readlink($fs_dir);
        # For relative links, need to munge the filename
        if (substr($linkdir,0,1) ne '/') {
            $fs_dir =~ s|[^/]+/*$||;
            $fs_dir .= "/$linkdir";
            while ($fs_dir =~ s|/+[^\./]+/+\.\.||g) { 1 };
        }
    }
 
    my @dirlist = ( $fs_dir );
    my @filelist = ();
    while (@dirlist) {
        my $dir = shift @dirlist;
        my $rel_dir = $dir;
        $rel_dir =~ s|^$fs_dir||;
        $rel_dir = substr($rel_dir,1) if substr($rel_dir, 0, 1) eq '/';
        if (! -d "$dir") {
            warning("Error: directory $web_dir/$rel_dir not found.", $errors);
            next;
        }
        if (!opendir DIR, "$dir") {
            warning("Error: unable to open directory $web_dir/$rel_dir.", $errors);
            next;
        }
        my @list = reverse sort(CORE::readdir(DIR));
        for (@list) {
            next if m/^\./;
            # If recursive (tree scope) add directories to dirlist
            if (-d "$dir/$_") {
                # Add to dirlist if recursive && ! source-control dir
                unshift @dirlist, "$dir/$_" if $recursive && ! m/^(CVS|RCS|{arch})$/;
                next;
            }
            # TODO: need a check for locked files, presumably omitted
            push @filelist, (($rel_dir) ? "$rel_dir/$_" : $_);
        }
        closedir DIR;
    }
 
    if (@filelist) {
        # If this is a directory link, publish the link itself as well
        if ($fs_link) {
            $fs_link =~ m|^(.*/)([^/]+/*$)|;
            # Copy the publishargs and explicitly turn off copylinks
            # my $fs_arg = { map { $_ => $arg{$_} } keys %arg };
            # $fs_arg->{copylinks} = 0;
            my $pubList = $results->{pubList};
            $results->{pubList} = [];
            publish($req, [ $2 ], $1, $results, %arg) if $1 && $2;
            # Fake a null results filename for the link itself
            push @$pubList, '';
            $results->{pubList} = $pubList;
        }
        # Call publish with the filelist
        return publish($req, \@filelist, $fs_dir, $results, %arg);
    }
}
 
1;
