# tag: ewok locking facade class

package Ewok::Locker;

use Carp;
use Ewok::Locker::DBM;
use Ewok::Config qw($EWOK_LOCKTYPE $EWOK_LOCK_TIMEOUT);
use strict;

my %type = map { $_ => 1 } qw(DBM);

sub init
{
    my $self = shift;
    my %arg = @_;
    my $type = $arg{type};
    $type ||= $EWOK_LOCKTYPE;
    croak "invalid type argument '$type' to Ewok::Locker->new()" 
        if ! exists $type{$type};

    # Create a new $type locker and cache
    $self->{locker} = "Ewok::Locker::$type"->new(
        timeout => $arg{timeout} || $EWOK_LOCK_TIMEOUT);

    return $self;
}

sub new 
{
    my $class = shift;
    my $self = {};
    bless $self, $class;
    return $self->init(@_);
}

#
# Create a lock on $file for $user. Fails if file is already locked (by 
#   someone else) and (if $LOCK_TIMEOUT is non-zero) the lock has not timed out, 
# Returns true on success, undef on error, and errors in $errors arrayref 
#   (if provided).
#
sub lock
{
    my $self = shift;
    my ($file, $user, $args) = @_;
    $args = {} if ref $args ne 'HASH';
    my $errors = $args->{errors};
    my $altfile = $args->{altfile};
    my $op = $args->{op};
    $errors  ||= [];
    $altfile ||= $file;
    $op ||= 'lock';
    $file = substr($file, 0, length($file)-1) 
        if substr($file, length($file)-1, 1) eq '/';

    croak "[lock] no lock object found" if ! ref $self->{locker};
    croak "[lock] no file parameter given" if ! $file;
    croak "[lock] no user parameter given" if ! $user;
    
    # Fail if 'lock' unsuccessful
    my $who;
    if (! $self->{locker}->lock($file, $user, { who => \$who })) {
        if ($who) {
            info("Error: cannot $op - locking file '$altfile' failed, locked by '$who'.", $errors);
        }
        else {
            info("Error: cannot $op - locking file '$altfile' failed.", $errors);
        }
        return undef;
    }

    return 1;
}

#
# Unlock $file for $user. Fails if locked by someone else.
# Returns true on success, 0 if the file wasn't locked, and undef on error, 
#   (with errors in $errors arrayref, if provided).
#
sub unlock
{
    my $self = shift;
    my ($file, $user, $args) = @_;
    $args = {} if ref $args ne 'HASH';
    my $errors = $$args{errors};
    my $altfile = $$args{altfile};
    my $op = $args->{op};
    $errors  ||= [];
    $altfile ||= $file;
    $op ||= 'unlock';
    $file = substr($file, 0, length($file)-1) 
        if substr($file, length($file)-1, 1) eq '/';

    croak "[lock] no lock object found" if ! ref $self->{locker};
    croak "[lock] no file parameter given" if ! $file;
    croak "[lock] no user parameter given" if ! $user;
    
    # Fail if 'lock' unsuccessful
    my $who;
    if (! $self->{locker}->unlock($file, $user, { who => \$who })) {
        if ($who) {
            info("Error: cannot $op - unlocking file '$altfile' failed, locked by '$who'.", $errors);
        }
        else {
            info("Error: cannot $op - unlocking file '$altfile' failed.", $errors);
        }
        return undef;
    }

    return 1;
}

1;

__END__

=head1 NAME

Ewok::Locker - ewok Locker class


=head1 SYNOPSIS

    use Ewok::Locker;

    # Create a new locker object
    $l = Ewok::Locker->new();
    # Or:
    $l = Ewok::Locker->new(type => 'DBM', timeout => 30);

    # Lock and unlock
    $l->lock($file, $user) || die "lock failed";
    $l->unlock($file, $user) || die "unlock failed";

    # Optional arguments
    $l->lock($file, $user, { errors => \@errors, 
        altfile => '/foo/bar.html', op => 'delete' });
    $l->unlock($file, $user, { errors => \@errors, 
        altfile => '/foo/bar.html', op => 'delete' });


=head1 DESCRIPTION

Ewok::Locker is a locker object that controls access to named objects.
It provides a lightweight cooperative locking scheme to minimise 
contention for shared objects in a browser-based web environment. It
does not enforce locking and it is not intended to be highly secure.

Objects are locked with a username and a timestamp. Attempts to lock or
unlock an object locked by someone else fail unless the lock has timed 
out. Users can nominate the timeout period to use (in minutes) when 
they instantiate the object. Locking an object for which the current
user already has a lock simply updates the lock timestamp.


=head1 BUGS

None known.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2001-2002, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
