# tag: ewok logger class using the apache error log

package Ewok::Logger::Apache;

use Apache;
use Apache::Log qw(DEBUG ERROR);
use Ewok::Config qw($EWOK_ERROR_LEVEL $EWOK_DEBUG_MODULES);
use Ewok::Logger::Null;

@ISA = qw(Ewok::Logger);

use strict;

my $singleton = undef;
my $level = {
    debug => 7,
    info => 6,
    'warn' => 4,
    error => 3,
    fatal => 2,
};

# Mapping specific subroutines to module subcategories
my $DEBUG_MAP = {
    'Ewok::File::publish' => 'Ewok::File::publish',
    'Ewok::File::unpublish' => 'Ewok::File::publish',
    'Ewok::File::publish_embperl' => 'Ewok::File::publish',
};

sub debug 
{
    return if $EWOK_ERROR_LEVEL ne 'debug';
    my $self = shift;
    
    # Debug statements only fire for 'modules' in $EWOK_DEBUG_MODULES
    # Subroutines can be mapped to fake or further subcategorised modules
    #   in $DEBUG_MAP, otherwise the real module is used (what TODO about 
    #   non-modules?)
    my $msg;
    if (ref $EWOK_DEBUG_MODULES eq 'ARRAY' && @$EWOK_DEBUG_MODULES) {
        my $caller = (caller 2)[3]; 
        my $module = $DEBUG_MAP->{$caller};
        ($module = $caller) =~ s/::[^:]+$// if ! $module;
        for (@$EWOK_DEBUG_MODULES) {
            $msg = join(' ', @_) if $module =~ m/^$_/;
        }
    }

    $self->{logger}->debug($msg) if $msg;
}

sub info 
{
    return if $EWOK_ERROR_LEVEL ne 'debug' && $EWOK_ERROR_LEVEL ne 'info';
    my $self = shift;
    my $msg = join(' ', @_);
    $self->{logger}->info($msg);
}

sub warning
{
    my $self = shift;
    my $msg = join(' ', @_);
    $self->{logger}->warn($msg);
}

sub error 
{
    my $self = shift;
    my $msg = join(' ', @_);
    $self->{logger}->error($msg);
}

sub fatal
{
    my $self = shift;
    my $msg = join(' ', @_);
    $self->{logger}->crit($msg);
}

sub new 
{
    my $class = shift;
    my ($category) = @_;
    if (! defined $singleton) {
        # Trap non-mod_perl environments
        my $r = eval { Apache->request() };
        return Ewok::Logger::Null->new() if ! defined $r;

        # Add apache log object to singleton
        $singleton = { logger => $r->log() };
        bless $singleton, $class;

        # Set loglevel
        my $s = $r->server();
        if (my $loglevel = $level->{$Ewok::Config::EWOK_ERROR_LEVEL}) {
            $s->loglevel($loglevel);
        }
        else {
            $r->log_error("Missing or invalid \$Ewok::Config::EWOK_ERROR_LEVEL: $Ewok::Config::EWOK_ERROR_LEVEL");
        }
    }
    return $singleton;
}

1;

__END__

=head1 NAME

Ewok::Logger::Apache - ewok logger class using the apache error log


=head1 SYNOPSIS

    use Ewok::Logger;



=head1 DESCRIPTION



=head1 BUGS

None known.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2001-2002, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
