# tag: ewok metadata module

package Ewok::Meta;

use Carp;
use File::Copy;
use Ewok::Config qw($EWOK_METADIR $EWOK_DIRI_OBJECT $EWOK_DIRN_OBJECT);
use Ewok::Util qw(:load &makedir);
use Ewok::Metamap qw(:all);
# use warnings;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(&mvMeta &getMeta &delMeta &setMeta &metaglob);
%EXPORT_TAGS = (
    all => [@EXPORT_OK],
);

use strict;

#
# Strip any inheriting directory metadata elements from the given $meta, 
#   returning the stripped elements in a new hashref
#
sub stripInherit
{
    my ($meta, $map) = @_;
    my $inherit = {};
    die "[stripInherit] meta argument not hashref" if ref $meta ne 'HASH';
    die "[stripInherit] map argument not hashref" if ref $map ne 'HASH';

    # delete $meta->{inherited};
    for my $key (keys %$meta) {
        if (ref $map->{$key} eq 'HASH' && $map->{$key}->{inherit}) {
            $inherit->{$key} = $meta->{$key};
            delete $meta->{$key};
        }
    }
    return $inherit;
}

#
# Derive metadata type, directory, and file names from the given $object.
#
sub mapObject
{
    my ($object, $arg) = @_;
    my $type = $arg->{type};
    croak "'$object' is not a directory" if $type eq 'dir' && ! -d $object;
    $type ||= ($object !~ m|/|) ? 'user' : -d $object ? 'dir' : 'file';

    my ($dir, $file);
    if ($type eq 'file') {
        ($dir, $file) = ($object =~ m|^(.*?)([^/]*)$|);
        $dir .= '.meta';
        $file .= '.meta';
    }
    elsif ($type eq 'dir') {
        $object = substr($object,0,-1) if substr($object,-1) eq '/';
        $dir .= "$object/.meta";
        $file = $EWOK_DIRN_OBJECT;
    }
    else {
        $dir = "$EWOK_METADIR/$type";
        $file = $object;
        croak "invalid metadata type '$type'" if ! -d $dir;
    }
    $dir = substr($dir,0,-1) if substr($dir,-1) eq '/';

    return $type, $dir, $file;
}

#
# Move the metadata associated with $objectFrom to be associated with $objectTo
#
sub mvMeta
{
    my ($fs_dir, $objectFrom, $objectTo) = @_;
    if (-d $fs_dir && -f "$fs_dir/.meta/$objectFrom.meta") {
        move("$fs_dir/.meta/$objectFrom.meta", "$fs_dir/.meta/$objectTo.meta");
    }
}

#
# Return the metadata hashref for $object, overriding any undefined metadata 
#   from the $defaults hashref, if given.
# Croaks on error.
#
sub getMeta
{
    my ($object, $arg) = @_;
    my $inherit = $arg->{inherit};
    my $defaults = $arg->{defaults} || {};
    croak "[getMeta] no object found" if ! $object;

    my ($type, $dir, $file) = mapObject(@_);
    return {} if ($type eq 'file') && ! -d $dir;

    my $meta = { %$defaults };
    if (-r "$dir/$file" && ($type ne 'dir' || $inherit ne 'I')) {
        my $meta2 = fileLoad($file, $dir, { 'eval' => 1, name => 'meta' }) || {};
        if (ref $meta2 eq 'HASH') {
            for (keys %$meta2) {
                $meta->{$_} = $meta2->{$_} if defined $meta2->{$_};
            }
        }
    }

    # If given, add defaults currently undefined in %meta
    if (ref $defaults eq 'HASH' && keys %$defaults) {
        my ($map) = getMap($object);
        for my $key (keys %$map) {
            $meta->{$key} = $defaults->{$key}
                if ! defined $meta->{$key} && defined $defaults->{$key};
        }
    }

    # For directories, add inherited directory metadata
    if ($type eq 'dir' && $inherit ne 'N') {
        my $meta2 = fileLoad($EWOK_DIRI_OBJECT, $object, 
            { 'eval' => 1, name => 'meta', inherit => 1, notfound => {} });
        if (ref $meta2 eq 'HASH') {
            for (keys %$meta2) {
                $meta->{$_} = $meta2->{$_} if defined $meta2->{$_};
            }
        }
    }

    return $meta;
}

#
# Removes all metadata for $object.
#
sub delMeta
{
    my ($object) = @_;
    croak "[delMeta] no object found" if ! $object;
    my ($type, $dir, $file) = mapObject($object);
    return unlink "$dir/$file" if -f "$dir/$file";
    return 1;
}

#
# Replace the metadata for $object with the values in %$meta.
# Returns true on success, undef (or croaks) on error.
#
sub setMeta
{
    my ($meta, $object, $arg) = @_;
    my $create = $arg->{create};
    my $errors = $arg->{errors} || [];
    my $map = $arg->{map};
    croak "[setMeta] no object found" if ! $object;
    croak "[setMeta] no metadata found" if ! $meta && ! $create;

    my ($type, $dir, $file) = mapObject($object, $arg);

    # Load metadata map for this object
    my $maptype = $type eq 'file' ? $object : $type;
    $map = getMap($maptype, $arg) if ! $map;
    croak "[setMeta] unable to load metamap for '$object'" if ! $map;

    # If this is a create, copy defaults from $map into $meta (except dir 'inherit')
    my $current;
    if ($create) {
        for my $mkey (keys %$map) {
            next if ref $map->{$mkey} ne 'HASH';
            my $default = $map->{$mkey}->{default};
            $meta->{$mkey} = $default 
                if $default ne '' && ! $map->{$mkey}->{inherit}; 
        }
    }
    # If not a create we're going to want to keep current 'FIXED' metadata
    else {
        # TODO: there is a race condition between the read here and the 
        #   write in the first fileSave below, although in practice it 
        #   probably doesn't matter since we're only ever using 'FIXED'
        #   (i.e. 'create') metadata from the read, which shouldn't change
        $current = fileLoad($file, $dir, { eval => 1, name => 'meta', 
            notfound => {} });
    }

    for my $key (keys %$meta) {
        # Remove any keys not in the metamap
        if (! $map->{$key}) {
            delete $meta->{$key};
        }
        # Unless $create is set, do a merge, keeping all fixed fields
        elsif (! $create && $map->{$key}->{fieldtype} eq 'FIXED') {
            $meta->{$key} = $current->{$key} if defined $current->{$key};
        }
    }

    # Strip inherited directory metadata if $type eq 'dir'
    my $inherit = stripInherit($meta, $map) if $type eq 'dir';

    # Save metadata
    my $rc = 1;
    if (keys %$meta) {
        if (($type eq 'file' || $type eq 'dir') && ! -d $dir) {
            if (! makedir($dir, { errors => $errors })) {
                $rc = 0;
            }
        }
        $rc = fileSave($meta, $file, $dir, { name => 'meta' }) if $rc;
    }

    # Save inherited directory metadata as well
    my $rc2 = 1;
    $rc2 = fileSave($inherit, $Ewok::Config::EWOK_DIRI_OBJECT, $object,
        { name => 'meta' }) if $type eq 'dir' && keys %$inherit;

    return $rc && $rc2;;
}

#
# Return an arrayref of files in $fs_dir whose filenames match $pattern 
#   subject to the $start and $max selectors. All other items in the arg
#   hash (apart from the $ascend sort flag) are taken to be metadata values,
#   all of which the file must match to be selected. (Values may be coderefs,
#   which are executed with the given metadata value).
# In array context also returns a hashref of metadata for each file, and a 
#   $more flag if further files matching the criteria exist.
#
sub metaglob ($$$;$)
{
    my ($fs_dir, $pattern, $arg) = @_;
    my $ascend = $arg->{ascend};
    my $start = $arg->{start};
    my $max = $arg->{max};
    delete $arg->{ascend};
    delete $arg->{start};
    delete $arg->{max};
    # Untaint
    ($fs_dir) = ($fs_dir =~ m|^\s*(.*?)\s*$|);
    return undef if ! -d $fs_dir || ! chdir $fs_dir;

    my @tmp = glob $pattern;
    # Sort in descending order (on filenames) by default
    @tmp = sort { ($ascend) ? $a cmp $b : $b cmp $a } @tmp;
    my @set = ();
    my $cnt = 0;
    my $meta = {};
    for my $f (@tmp) {
      $meta->{$f} = getMeta("$fs_dir/$f");
      my $match = 1;
      # Check for metadata matches - must match all
      for my $m (keys %$arg) {
          if (ref $arg->{$m} eq 'CODE') {
              $match = &{$arg->{$m}}($meta->{$f}->{$m});
          }
          else {
              $match = 0 if $arg->{$m} ne $meta->{$f}->{$m};
          }
          if (! $match) {
              delete $meta->{$f};
              last;
          }
      }
      if ($match) {
        next if $start > $cnt++;
        push @set, $f;
        if (scalar(@set) > $max) {
          last;
        }
      }
    }
    my $more = 0;
    if (scalar(@set) > $max) {
      delete $meta->{pop @set};
      $more = 1;
    }
    return (wantarray) ? (\@set, $meta, $more) : \@set;
}

1;
