# tag: ewok metadata map module

package Ewok::Metamap;

use Ewok::Config qw($EWOK_METAMAP_OBJECT $EWOK_HOME);
use Ewok::Util qw(:all);
use Ewok::Authz qw(:common);

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(default datatype fieldtype priv inherit 
                    SYSTEM MANDATORY OPTIONAL FIXED 
                    TEXT TEXTLONG INTEGER FLOAT BOOLEAN
                    DATE DATETIME USERNAME EMAIL URL
                    &getMapItemProp &setMapItemProp &getMap &checkType);
%EXPORT_TAGS = (
    all => [@EXPORT_OK],
);

use strict;

#
# Metamap properties and values
#
# A metadata map (by default in 'metamap.epl') is a hashref mapping objects
#   and object types to metadata item definitions. Objects can be any of 
#   'dir', 'file', or 'user', or a filetype regular expression. The getMap
#   subroutine returns the appropriate submap for the given object.
#
# A submap, containing the metadata item definitions for a particular object
#   or set of filetypes, is an arrayref containing metadata item => property
#   hashref mappings. Metadata property hashrefs can contain the following 
#   keys:
#     - default         item default value (literal or 'SYSTEM')
#     - datatype        item datatype
#     - fieldtype       item fieldtype (MANDATORY, OPTIONAL, or FIXED)
#     - hidden          Y/NULL, whether item is hidden from display (def=NULL)
#     - priv            privilege level required to modify (def=NULL)
#     - inherit         for directory metadata, whether this item is inherited
#                         by descendant items (files/directories)
#
# 'SYSTEM' defaults must be supported by the setSystemDefaults subroutine 
#

my @PROPLIST = qw(default datatype fieldtype priv inherit);
use constant default => 'default';
use constant datatype => 'datatype';
use constant fieldtype => 'fieldtype';
use constant hidden => 'hidden';
use constant priv => 'priv';
use constant inherit => 'inherit';

# Special default value
use constant SYSTEM    => 'SYSTEM';

# Datatypes
use constant TEXT     => 'TEXT';       # One-line textbox in HTML
use constant TEXTLONG => 'TEXTLONG';   # Multi-line textarea in HTML
use constant TEXTWORD => 'TEXTWORD';   # Textbox, no spaces allowed
use constant INTEGER  => 'INTEGER';
use constant FLOAT    => 'FLOAT';
use constant BOOLEAN  => 'BOOLEAN';
use constant DATE     => 'DATE';
use constant DATETIME => 'DATETIME';
use constant USERNAME => 'USERNAME';
use constant PASSWORD => 'PASSWORD';   # Password field
use constant EMAIL    => 'EMAIL';
use constant URL      => 'URL';
use constant ENUM     => 'ENUM';

# Field properties
#   MANDATORY: required
#   OPTIONAL:  optional
#   FIXED:     fixed (cannot be modified, therefore requires default)
use constant MANDATORY => 'MANDATORY';
use constant OPTIONAL  => 'OPTIONAL';
use constant FIXED     => 'FIXED';


# Accessor routine for metamap properties (primarily for implementation 
#   hiding). If $prop is set, returns the property requested, otherwise 
#   the property list hashref, or undef on error.
#
sub getMapItemProp
{
    my ($map, $item, $prop) = @_; 

    # Error conditions
    return undef if ref $map ne 'HASH';
    return undef if defined $prop && ! grep /^$prop$/, @PROPLIST;
    return undef if ! exists $map->{$item};
    return undef if ref $map->{$item} ne 'HASH';

    if (defined $prop) {
        return $map->{$item}->{$prop};
    }
    else {
        return $map->{$item};
    }
}

#
# Return a list of items containing the given $prop and $value
#
sub getMapItems
{
    my ($map, $prop, $value) = @_;
    my @items = ();

    # If $map is a scalar do a getMap on it
    $map = getMap($map) if ! ref $map;

    # Error conditions
    return undef if ref $map ne 'HASH';
    return undef if defined $prop && ! grep /^$prop$/, @PROPLIST;

    for my $item (%$map) {
        next if $item eq 'TYPE';
        push @items, $item if $map->{$item}->{$prop} eq $value;
    }

    return @items;
}

#
# Mutator routine for metamap properties (primarily for implementation 
#   hiding). Returns true on success, or undef on error.
#
sub setMapItemProp
{
    my ($map, $item, $prop, $value) = @_;

    # Error conditions
    return undef if ! grep /^$prop$/, qw(default datatype fieldtype);
    return undef if ! exists $map->{$item};
    return undef if ref $map->{$item} ne 'HASH';

    $map->{$item}->{$prop} = $value;
    return 1;
}

#
# Sets up SYSTEM defaults in the given metamap. Create defaults are set only
#   at object creation time, i.e. if the $create flag is set.
# No return values (void).
#
sub setSystemDefaults
{
    my ($map, $arg) = @_;
    my $create = $arg->{create};
    my $user = $arg->{user} || $ENV{REMOTE_USER};
    my $modify_dtm;

    my @create = qw(creator create_dtm);
    my @mon = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
    my @now = localtime;
    my $value;
    for my $key (keys %$map) {
        next if getMapItemProp($map, $key, default) ne SYSTEM;

        # Standard defaults
        if ($key eq 'owner') {
            $value = $user || '';
            setMapItemProp($map, $key, default, $value);
            next;
        }

        # Create-only defaults
        elsif (grep /^$key$/, @create) {
            next if ! $create;
            if ($key eq 'creator') {
                $value = $user || '';
                setMapItemProp($map, $key, default, $value);
                next;
            }
            elsif ($key eq 'create_dtm') {
                $value = "$now[3] $mon[$now[4]] " . ($now[5]+1900) .
                    sprintf(" %2d:%02d:%02d", $now[2], $now[1], $now[0]);
                setMapItemProp($map, $key, default, $value);
                next;
            }
        }

        # Anything else
        else {
            $value = 'UNKNOWN';
            setMapItemProp($map, $key, default, $value);
            next;
        }
    }

#   # Special create_dtm hack - set to modify_dtm.
#   # Should never be required except for files not created through ewok.
#   if (exists $map->{create_dtm} && $modify_dtm) {
#       setMapItemProp($map, 'create_dtm', default, $modify_dtm);
#   }
}

#
# Finds the metadata map for $type. If $type begins with a '/', it is taken 
#   to be a filename and we search for the appropriate map for that file.
# Returns two items: a hashref to the map itself and an (ordered) arrayref of
#   the map keywords. On error, the map hashref is returned as undef.
#
sub getMap
{
    my ($type, $arg) = @_;
    my $create = $arg->{create};
    my $ar = $arg->{ar};
    die "[Metamap::getMap] no type found" if ! $type;

    # Use $type as $fs_file if begins with '/'
    my $fs_file = $type if $type =~ m|^\s*/|;
    my $context = $fs_file || "$EWOK_HOME/htdocs";

    # Load the metamap for $fs_file
    my $metamap = fileLoad($EWOK_METAMAP_OBJECT, $context, 
        { 'eval' => 1, name => 'metamap', inherit => 1, ar => $ar }) || {};
    # If any of the default entries are missing, load them from $EWOK_HOME
    if (! exists $metamap->{dir} || ! exists $metamap->{file} || ! exists $metamap->{user}) {
        my $basemap = fileLoad($EWOK_METAMAP_OBJECT, "$EWOK_HOME/htdocs", 
            { 'eval' => 1, name => 'metamap', inherit => 1, cache => 1 }) || {};
        $metamap->{user} = $basemap->{user} if ! exists $metamap->{user};
        $metamap->{dir}  = $basemap->{dir}  if ! exists $metamap->{dir};
        $metamap->{file} = $basemap->{file} if ! exists $metamap->{file};
    }

    # For directories use the 'dir' entry
    if ($fs_file && -d $fs_file) {
        $type = 'dir';
    }

    # Find files by filetype
    elsif ($fs_file) {
        $type = undef;
        for my $pattern (keys %$metamap) {
            next if $pattern eq 'file';
            if ($fs_file =~ m/$pattern$/) {
                $type = $pattern;
                last;
            }
        }
        # Default to 'file' if no matches
        $type ||= 'file';
    }

    # Load the map for the given key
    my @keys = ();
    if (defined $type && exists $metamap->{$type}) {
        my $mapArray = $metamap->{$type};
        if (ref $mapArray eq 'ARRAY') {
            my $exists = {};
            # Setup keys array (ignore 'create' elts unless $create set)
            if (wantarray) {
                for (my $i = 0; $i < $#$mapArray; $i += 2) {
                    push @keys, $mapArray->[$i]
                        if $create || ! $mapArray->[$i+1]->{create};
                    $exists->{omit}   ||= $mapArray->[$i+1]->{omit};
                    $exists->{create} ||= $mapArray->[$i+1]->{create};
                }
            }

            # Remove 'create' metadata from map if $create flag not set
            # my $map = { @{deepcopy($mapArray)} };
            my $map = { @$mapArray };
            if (! $create) {
                for my $m (keys %$map) {
                    delete $map->{$m} if $map->{$m}->{create};
                }
            }
            # Insert 'TYPE' element
            $map->{TYPE} = $type;

            # Populate any SYSTEM defaults in the map
            setSystemDefaults($map, $arg);

            return ((wantarray) ? ( $map, \@keys, $exists ) : $map);
        }
    }

    return undef;
}

#
# Check that $value is an acceptable value to be a $type.
# Returns true or false appropriately, or undef for some other error.
#
sub checkType
{
    my ($name, $value, $type, $error, $value2) = @_;
    return undef if ref $error ne 'SCALAR';
    $$error = '';

    # No type checking for TEXT or TEXTLONG
    return 1 if $type eq TEXT || $type eq TEXTLONG;

    # Strip leading and trailing space
    $type =~ s/^\s+//o;
    $type =~ s/\s+$//o;

    # Do the appropriate check
    if ($type !~ m/\|/o) {
        if ($type eq INTEGER) {
            return 1 if ($value =~ m/^[+-]?\d+$/o);
            $$error = "Error: $name '$value' is not a valid \L$type.";
            return 0;
        }
        elsif ($type eq FLOAT) {
            return 1 if ($value =~ m/^[+-]?\d*(\.\d+)?$/o);
            $$error = "Error: $name '$value' is not a valid \L$type.";
            return 0;
        }
        elsif ($type eq DATE || $type eq DATETIME) {
            if (defined eval { require Date::Parse }) {
                my $time = str2time($value);
                return 1 if $time;
            }
            elsif (defined eval { require Date::Manip }) {
                Date_Init("DateFormat=non-US");
                my $date = ParseDate($value);
                return 1 if $date;
            }
            else {
                # TODO: revisit what to do here
                return 1;
            }
            $$error = "Error: $name '$value' is not a valid date/time.";
            return 0;
        }
        elsif ($type eq USERNAME) {
            return 1 if ($value =~ m/^\S+$/o);
            $$error = "Error: $name '$value' is not a valid \L$type.";
            return 0;
        }
        elsif ($type eq PASSWORD) {
            return 1 if ($value && $value eq $value2);
            $$error = "Error: passwords do not match.";
            return 0;
        }
        elsif ($type eq EMAIL) {
            return 1 if ($value =~ m/^[\w\.\-\+]+\@[\w\.\-\+]+/o);
            $$error = "Error: $name '$value' is not a valid \L$type.";
            return 0;
        }
        elsif ($type eq URL) {
            return 1 if ($value =~ m|^https?://|o);
            $$error = "Error: $name '$value' is not a valid \L$type.";
            return 0;
        }
        elsif ($type eq BOOLEAN) {
            return 1 if ($value eq '1' || $value eq '0');
            $$error = "Error: $name '$value' is not a valid \L$type.";
            return 0;
        }
        else {
            $$error = "Error: unknown datatype '$type'.";
            return undef;
        }
    }
    # If select list
    else {
        my @list = split /\s*\|\s*/, $type;
        $value =~ s/^\s+//o;
        $value =~ s/\s+$//o;
        return 1 if grep /^$value$/, @list;
        $$error = "Error: $name '$value' is not a recognised value ($type).";
        return undef;
    }

    $$error = "Error: unknown error.";
    return 0;  # Test only
}

1;
