# tag: ewok parsing subroutines module

package Ewok::Parse;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = ();
@EXPORT_OK = qw(&parseFileIni &parseFileSep);
$VERSION = 0.2;

use strict;

#
# Parse the given file, treating it as a MS-style ini file, with [ Sections ]
#   heading name = value pairs. Return a hashref of sections, pointing to
#   hashrefs containing the name-value pairs. If the 'order' option is set,
#   also return an ordered array of section names.
#
sub parseFileIni ($;$) 
{
    my $filename = shift;
    my $options = shift;
    my $order = exists $options->{order};
    my $duplicates = $options->{duplicates} || 'concat';
    my $duplicateSep = $options->{duplicateSep} || ',';
    my %result = ();
    my @ordering = ();
    open FILE, "<", $filename or return undef;
    my $section = '';
    my $file;
    {
        # $/ can't be set to a regex, which means handling both unix and dos
        # eol conventions via <FILE> is non-trivial - hence slurp and split
        local $/ = undef;
        $file = <FILE>;
    }
    for (split /\r?\n/, $file) {
        # End on special __END__ token
        last if /^__END__/;
        # Ignore comments and empty lines
        next if /^#/ || /^\s*$/;
        # Sections
        if (/^\s*\[\s*(.*?)\s*\]/) {
            $section = $1;
#           $section =~ s/\s+//go;      # Strip spaces
            $result{$section} = {};
            push @ordering, $section if $order;
        } 
        # Name-value pairs
        elsif ($section && /^\s*(.+?)\s*=\s*(.*?)\s*$/) {
            my $sectionHash = $result{$section};
            if ($duplicates eq 'concat') {
                if (defined $sectionHash->{$1}) {
                    $sectionHash->{$1} .= $duplicateSep . $2;
                }
                else {
                    $sectionHash->{$1} = $2;
                }
            }
            else {
                if (exists $sectionHash->{$1}) {
                    for my $i (2 .. 99) {
                        next if exists $sectionHash->{"$1$i"};
                        $sectionHash->{"$1$i"} .= $2;
                        last;
                    }
                }
                else {
                    $sectionHash->{$1} = $2;
                }
            }
        }
    }
    if ($order) {
      return \%result, \@ordering;
    }
    else {
      return \%result;
    }
}

#
# Parse the given file, splitting by line and then by the given separator,
#   returning an arrayref of arrayrefs, or undef on error
#
sub parseFileSep ($;$$) 
{
    my $filename = shift;
    my $delim = shift;
    my $maxFields = shift;
    # Escape regex special chars
    $delim =~ s#([\|\$\^\\\/])#\\$1#og;
    $delim ||= '\|';
    my @a = ();
    open FILE, "<", $filename 
        or return undef, "parseFileSep cannot open $filename";
    while (<FILE>) {
        chomp;
        # Ignore comments and empty lines
        next if /^#/ || /^\s*$/;
        # Split, deleting surrounding whitespace
        if ($maxFields) {
            push @a, [ split /\s*$delim\s*/, $_, $maxFields ];
        } 
        else {
            push @a, [ split /\s*$delim\s*/, $_ ];
        }
    }
    close FILE;
    return \@a;
}

1;
