# tag: ewok text markup module

package Ewok::TextMarkup;

require Exporter;
use Carp;
use Ewok::Config qw($EWOK_AUTO_CONTENTS_STRING $EWOK_MARKUP_DOCTYPE);

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
    &markup
    &parse_headings
);
%EXPORT_TAGS = (
    'all' => [@EXPORT_OK],
);

use strict;

#
# HTML-specific markup to be done before any text-based markup 
#   (html heading anchoring)
#
sub markup_html1
{
    my %arg = @_;
    $_ = $arg{input};
    my $markup_heading_anchors = $arg{markup_heading_anchors};
    $markup_heading_anchors = 1 if ! defined $markup_heading_anchors;

    # Escape quotes in headings
    while (s|(<h([1-6])>.*?)"(.*</h\2>)|$1&quot;$3|img) { 1 }; #"
    if ($markup_heading_anchors) {
        # <hX>Foo</hX> => <hX>"Foo"=#foo</hX>
        s|(<h([1-6])[^>]*>)\s*(.+?)\s*(</h\2>)(\r?\n)?|$1"$3"=#\L$3\E$4\n|img;
    }

    return $_;
}

#
# Subroutine to markup lists and <pre> sections - procedural (hard via regexes)
#
sub markup_indented
{
    my $text = shift;
    my %arg = @_;
    my $markup_classes_styles = $arg{markup_classes_styles};
    my $output = '';
    my @indent = (-1);
    my @type = ();
    my ($newindent, $extraindent, $class, $style, $item);
    my $orderedMarker = qr/N[AaIi]?\./;
    my $unorderedMarker = qr/[-\*o\+]/;
    my $listMarker = qr/$orderedMarker|$unorderedMarker/;
    my $type = '';
    my $para = 0;
    # Iterate over lines
    for (split /\r?\n/, $text) {
        # Non-indented line - reset
        if (m/^\S/) {
            if ($type eq 'pre') {
                $output .= "</pre>\n\n";
            }
            elsif ($type) {
                $output .= "</p>" if $para > 1;
                $output .= "</li>\n";
                # Close off hanging lists at first non-space non-list-item
                if ($indent[0] > -1) {
                    while ($indent[0] > -1) {
                        $output .= ' ' x $indent[0] . "</$type>\n\n";
                        shift @indent;
                    }
                }
            }
            $type = '';
            $para = 0;
            $output .= "$_\n";
        }
        # Empty line
        elsif (m/^(\s*)$/) {
            # For lists, simply set $para flag at empty line
            if ($type && $type ne 'pre') {
                if ($para) {
                    $output .= "</p>";
                }
                $para = 1;
            }
            # Otherwise pass straight through
            else {
                $output .= "\n";
            }
        }
        # Indented line
        elsif (m/^(\s+)/) {
            # If in <pre> section, pass straight through
            if ($type eq 'pre') {
                $output .= "$_\n";
            }
            # Check for initial list marker
            elsif (m/^(\s+)($listMarker)(\s+)(.+?)\s*$/) {
                $newindent = length($1);
                my $marker = $2;
                $type = substr($marker,0,1) eq 'N' ? 'ol' : 'ul';
                $extraindent = $newindent + length($2) + length($3);
                $item = $4;
                $class = '';
                $style = '';

                # Setup marker type
                my $mtype = '';
                if ($type eq 'ol') {
                    $marker = substr($marker,1,1);
                    $mtype = qq( type="$marker") if $marker =~ m/[AaIi]/;
                }
                else {
                    if ($marker eq '*') {
                        $mtype = ' type="disc"';
                    }
                    elsif ($marker eq 'o') {
                        $mtype = ' type="circle"';
                    }
                    elsif ($marker eq '+') {
                        $mtype = ' type="square"';
                    }
                }

                # Remove class and style bits from $item if $markup_classes_styles
                if ($markup_classes_styles && $item =~ m/(\[\s*(\S.*?)\s*\]\s*)?({\s*(\S.*?)\s*}\s*)?(.+?)\s*$/) {
                    $class = qq( class="$2") if $2;
                    $style = qq( style="$4") if $4;
                    $item = $5;
                }

                # Add new list item
                if ($newindent > $indent[0]) {
                    # Increased indentation - begin new list
                    $output .= "</p>" if $para > 1;
                    $output .= "</li>\n" if $indent[0] > -1;
                    unshift @indent, $newindent;
                    unshift @type, "$type:$marker";
                    $output .= ' ' x $indent[0] . "<$type$class$style>\n";
                    $output .= ' ' x $indent[0] . "<li$mtype>$item";
                }
                else {
                    # Terminate current item
                    $output .= "</p>" if $para > 1;
                    $output .= "</li>\n";
                    if ($newindent < $indent[0]) {
                        # Decreased indentation - terminate outstanding list(s)
                        while ($newindent < $indent[0]) {
                            my ($prev) = split /:/, shift @type;
                            $output .= ' ' x $indent[0] . "</$prev>\n";
                            shift @indent;
                        }
                    }
                    if ("$type:$marker" ne $type[0]) {
                        # Type or marker changed - begin a new list, same level
                        my ($prev) = split /:/, shift @type;
                        $output .= ' ' x $indent[0] . "</$prev>\n";
                        unshift @type, "$type:$marker";
                        $output .= ' ' x $indent[0] . "<$type$class$style>\n";
                    }
                    $output .= ' ' x $indent[0] . "<li$class$style$mtype>$item";
                }
                $para = 0;
            }
            # No list marker, not <pre> - check if in list at current indent level
            elsif ($type && m/^\s{$extraindent}(\S.*)/) {
                $output .= "\n" . ' ' x $indent[0];
                if ($para == 1) {
                    $para = 2;
                    $output .= "<p>";
                }
                $output .= $1;
            }
            # No list marker, not at current indent level - treat as new <pre>
            else {
                # If list, close off
                if ($type) {
                    $output .= "</p>" if $para > 1;
                    $output .= "</li>\n";
                    while ($indent[0] > -1) {
                        $output .= ' ' x $indent[0] . "</$type>\n";
                        shift @indent;
                        ($type) = split /:/, shift @type;
                    }
                }
                $para = 0;
                $output .= ' ' x $indent[0] if $indent[0] > -1;
                $type = 'pre';
                $output .= "<pre>\n";
                $output .= "$_\n";
            }
        }
    }

    # Check for hanging items at end of data
    if ($type) {
        if ($type eq 'pre') {
            $output .= "</pre>\n\n";
        }
        elsif ($indent[0] > -1) {
            $output .= "</p>" if $para > 1;
            $output .= "</li>\n";
            while ($indent[0] > -1) {
                $output .= ' ' x $indent[0] . "</$type>\n";
                shift @indent;
                ($type) = split /:/, shift @type;
            }
        }
    }

    return $output;
}

#
# Subroutine to markup text to html
#
sub markup_text
{
    my %arg = @_;
    local $_ = $arg{input};
    my $strict_paragraphs = $arg{strict_paragraphs};
    my $markup_escape_html = $arg{escape_html};
    my $markup_bold = $arg{markup_bold};
    my $markup_italics = $arg{markup_italics};
    my $markup_links = $arg{markup_links};
    my $markup_classes_styles = $arg{markup_classes_styles};
    my $markup_tables = $arg{markup_tables};
    my $markup_heading_anchors = $arg{markup_heading_anchors};
    $strict_paragraphs = 1 if ! defined $strict_paragraphs;
    $markup_bold = 1 if ! defined $markup_bold;
    $markup_italics = 1 if ! defined $markup_italics;
    $markup_links = 1 if ! defined $markup_links;
    $markup_escape_html = 1 if ! defined $markup_escape_html;
    $markup_classes_styles = 1 if ! defined $markup_classes_styles;
    $markup_tables = 1 if ! defined $markup_tables;
    $markup_heading_anchors = 1 if ! defined $markup_heading_anchors;

    # Escape html tags
    if ($markup_escape_html) {
        s|<|&lt;|g;
        s|>|&gt;|g;
    }

    # Markup list items and <pre> sections
    $_ = markup_indented($_, markup_classes_styles => $markup_classes_styles);

    # Markup tables (disabled)
    if (0) {
        my ($class, $style);
        if ($markup_classes_styles) {
            $class = '(?<!\\\\)\[\s*([^\]]*?)\s*\]\s*';
            $style = '(?<!\\\\){\s*([^\}]*?)\s*\}\s*';
        }
        my $text = '';
        for my $line (split /\r?\n/, $_) {
            # Begin-table tag (with support for classes and styles)
            if ($line =~ m/^[ \t]*($class)?($style)?($class)?($style)?(?<!\\)\|(.*)$/) {
                $line = '<table';
                if ($6 || $8) {
                    $line .= qq( class="$2") if $2;
                    $line .= qq( style="$4") if $4;
                }
                $line .= '><tr';
                if ($6 || $8) {
                    $line .= qq( class="$6") if $6;
                    $line .= qq( style="$8") if $8;
                }
                # If single set assume <tr> over <table>
                else {
                    $line .= qq( class="$2") if $2;
                    $line .= qq( style="$4") if $4;
                }
                $line .= "><td>$9";
            }
            $text .= "$line\n";
        }
        $_ = $text;
        # End-table tag 
        s:(?<!\\)\|\s*?$:</td></tr></table>:mg;
        # Between table cells
        s:(?<!\\)\|:</td><td>:g;
        # Markup <td> classes and styles
        s:<td([^>]*)>\s*$class:<td$1 class="$2">:g if $class;
        s:<td([^>]*)>\s*$style:<td$1 style="$2">:g if $style;
        # Remove extraneous <table> and </table> tags
        s|</table>(.*?\n.*?)<table[^>]*>|$1|g;
        # &nbsp; empty cells
        s|(<td[^>]*>)(</td>)|$1&nbsp;$2|g;
        # If <table> not deleted and single set, assume <table> over <tr>
        s|<table>\s*<tr(\s+class=".*?")?(\s+style=".*?")?\s*>|<table$1$2><tr>|g;
        # Border pseudoclass - translate a table [border] class to border="1"
        s|<table class="border"|<table border="1"|ig;
        # th pseudoclass - translate a tr [th] class to use <th> tags
        while (s|(<tr class="th".*?)(</?)td|$1$2th|im) { 1 };
        s|<tr class="th"|<tr|ig;
    }

    # Markup headings
    # Underlined headings - h1 to h3
    my $level = { '=' => 1, '-' => 2, '~' => 3 };
    s|^[ \t]*(\S.*)\n[ \t]*([-=~])[-=~]{2,}\s*?$|h$level->{$2}: $1|mg;
    if ($markup_heading_anchors) {
        # h1: Foo =#foofoo
        s|^[ \t]*(h[1-6]):[ \t]*(.+?)\s*(=#\S+)\s*$|<$1>"$2"$3</$1>\n|img;
        # h1: Foo
        s|^[ \t]*(h[1-6]):[ \t]*([^:].*?)\s*$|<$1>"$2"=#\L$2\E</$1>\n|img;   #"
        # h1:: Foo             #"
        s|^[ \t]*(h[1-6])::[ \t]*(.+?)\s*$|<$1>$2</$1>\n|img;
    }
    else {
        # h1: Foo             #"
        s|^[ \t]*(h[1-6])::?[ \t]*(.+?)\s*(=#\S+\s*)?$|<$1>$2</$1>\n|img;
    }

    # HRs
    s|^[-_]{4,}\s*?$|<hr>|mg;
    
    # Add an opening <p> at the very beginning
    s|^\s*|<p>|;
    # Add a closing <p> at the very end
    s|\s*$|</p>\n|;
    # Add paragraph tags at all empty lines
    s|\r?\n\s*(\r?\n)|</p>$1<p>|g;
    # Remove any final hanging <p> tag
    s|<p>\s*$||;
    # Remove any tags around headings
    s|^<p>(\s*"?\s*<h\d)|$1|img;   #"
    s|(</h\d>.*)</p>$|$1|img;
    # Remove any tags around lists + tables
    s|^<p>(\s*<table)|$1|img;
    s|(</table>\s*)</p>$|$1|img;
    if ($strict_paragraphs) {
        s|^<p>(\s*<(?![biu]\>))|$1|img;
        s|(?<!</[biu])>\s*</p>$|>|img;
    }
    # Add <br> tags on any short (<40 chars) lines (not ending in > or embperl [-+#\$!]\]
    s:^(\s*\S.{0,39}(?<!.>|[-+#\$!]\]))\s*?$:$1<br>:mg;
    # Remove <br> and <p> tags within <pre></pre> sections
    my $count = 0;
    my $new = '';
    for (split m/<pre>/i) {
        1 while s!(.*?)<p>(.*</pre>.*)!$1\n$2!isg;
        1 while s!(.*?)<(br|/p)>(.*</pre>.*)!$1$3!isg;
        $new .= '<pre>' if $count;
        $new .= $_;
        $count++;
    }
    $_ = $new;
    # Do some [class] and {style} checking for paragraphs
    if ($markup_classes_styles) {
        s!^<p>\s*\[\s*(\S.*?)\s*\]\s*!<p class="$1">!img;
        s!^<p([^>]*)>\s*{\s*(\S.*?)\s*\}\s*!<p$1 style="$2">!img;
    }

    # Add formatting tags (strong, em, etc)
    s|(?<!\\)\*(\S(.*?\S)??)(?<!\\)\*|<strong>$1</strong>|mg
        if $markup_bold;
    s|(?<!\\)__(\S(.*?\S)??)(?<!\\)__|<strong><em>$1</em></strong>|mg
        if $markup_bold && $markup_italics;
    s|(?<!\\)_(\S(.*?\S)??)(?<!\\)_|<em>$1</em>|mg
        if $markup_italics;

    # Unescape any escaped formatting tags
    s|\\([*_\|\[\{])|$1|mg;

    # Markup links
    if ($markup_links) {
        # Markup mailto links
        # Labelled mailto
        s!"([^"]+)":\s*(mailto:\w[-.+\w]+\@([-\w]+\.)+\w+(\?[^\s<]*[^\s<\.])?) #"
            !<a href="$2">$1</a>!imgx;
        # Labelled email
        s!"([^"]+)":\s*(\w[-.+\w]+\@([-\w]+\.)+\w+)  #"
            !<a href="mailto:$2">$1</a>!imgx;
        # Bare mailto
        s|(?<!\")(mailto:(\w[-.+\w]+\@([-\w]+\.)+\w+)(\?[^\s<]*[^\s<\.])?)  #"
            |<a href="$1">$2</a>|imgx;
        # Bare email in <>
        s! ([^:\w]) (<|&lt;) (\w[\-\.\+\w]+\@([-\w]+\.)+\w+) (>|&gt;)  #"
            !$1<a href="mailto:$3">&lt;$3&gt;</a> !igx;
        # Bare email
        s! ([^:;\w]) (\w[-.+\w]+\@([-\w]+\.) +\w+)  #"
            !$1<a href="mailto:$2">$2</a>!imgx;

        # Markup http links
        s!"([^"]+)":\s*(http://[^\s<]+[\w/][\w/\?=\+])    #"
                  !<a href="$2">$1</a>!imgx;
        s!"([^"]+)":\s*([^\s<]+[^\s<\.][\w/\?=\+])   #"
                  !<a href="$2">$1</a>!imgx;
        s|(?<!\")(http://[^\s<]+[^\s<\.][\w/\?=\+])     #"
                  |<a href="$1">$1</a>|imgx;
    }

    # Cleanup any (wrongly) formatted code sections
    if ($arg{allow_syntax} eq 'Text') {
        my $new = '';
        my $code_open = qr((\[?\[[-+#!\$]));
        my $code_type;
        for (split $code_open) {
            if ($_ =~ $code_open) {
                ($code_type) = m/\[([-+#!\$])/;
            }
            else {
                if ($code_type) {
                    my $b;
                    ($_, $b) = split /\Q$code_type\E\]/, $_, 2;
                    # <pre> tags on own line within code
                    1 while s:(.*?\n)</?pre>\r?\n(.*):$1$2:isg;
                    # <br> at $ within code
                    1 while s:(.*?)<br[ \t]*/?>(\r?\n.*):$1$2:isg;
                    # <p> at ^ within code
                    1 while s:(.*?\n)<p[ \t]*/?>(.*):$1$2:isg;
                    # </p> at $ within code
                    1 while s:(.*?)</p[ \t]*/?>(\r?\n.*):$1$2:isg;
                    $_ = $_ . "$code_type]" . $b;
                    # Reset $code_type
                    $code_type = '';
                }
            }
            $new .= $_;
        }
        $_ = $new;
        # <p> at ^ before code
        s:^<p[\t ]*/?>(\[[-+#\$!].*)$:$1:img;
        # </p> at $ after code
        s:^(.*[-+#\$!]\])\s*</p>$:$1:img;
    }

    # Format
    my $doctype = $EWOK_MARKUP_DOCTYPE;
    $doctype .= "\n" if $doctype;
    return <<EOD;
$doctype<html>
<body>
$_
</body>
</html>
EOD
}

#
# Parse the given html content for top-level headings. 
#
sub parse_headings
{
    my %arg = @_;
    local $_ = $arg{input};
    my $delete_initial_heading = $arg{delete_initial_heading};
    my $maxTags = 10;
    my ($lead, $stag, $etag);

    my $contents = '';
    $contents = $Ewok::Config::EWOK_AUTO_CONTENTS_STRING
        if $arg{auto_contents} && ! m!((<|&lt;)contents\b([^>]*)>)!is;

    # Find first h1 tag. This is somewhat nasty - need to try to limit 
    #   the number of tags we will search through to try and find our heading
    my ($h1, $h1_full);
    if ($delete_initial_heading || ref $arg{h1} eq 'SCALAR') {
        if (m|(<body.*?)(<h1\b[^>]*>)((?:<a name=[^>]+>)?(.*?)(?:</a>)?)(</h1.*?>)|is ||
            m|^(.*?)(<h1\b[^>]*>)((?:<a name=[^>]>)?(.*?)(?:</a>)?)(</h1.*?>)|is) {
            $lead = $1;
            $stag = $2;
            $h1_full = $3;
            $h1 = $4;
            $etag = $5;
        }
        # Limit no of tags between opening <body> (or beginning) and <h1>
        my $tags = ($lead =~ s|<|<|og);
        undef $h1 if $tags > $maxTags;

        # Do the h1 delete
        if ($h1) {
            if ($delete_initial_heading) {
                s|(<p>)?\s*$stag\Q$h1_full\E$etag(\s*</p>)?(\s*\r?\n)?|$contents|is;
            }
            elsif ($contents) {
                s|((<p>)?\s*$stag\Q$h1_full\E$etag(\s*</p>)?(\s*\r?\n)?)|$1$contents|is;
            }
        }
    }

    # Find first h2 tag. 
    my ($h2, $h2_full);
    if (ref $arg{h2} eq 'SCALAR') {
        if (m|(<body.*?)(<h2\b[^>]*>)((?:<a name=[^>]+>)?(.*?)(?:</a>)?)(</h2.*?>)|is ||
            m|^(.*?)(<h2\b[^>]*>)((?:<a name=[^>]>)?(.*?)(?:</a>)?)(</h2.*?>)|is) {
            $lead = $1;
            $stag = $2;
            $h2_full = $3;
            $h2 = $4;
            $etag = $5;
        }
        # Limit no of tags between opening <body> (or beginning) and <h2>
        my $tags = ($lead =~ s|<|<|og);
        undef $h2 if $tags > $maxTags;

        # Delete the h2 if no h1 found
        if ($h2 && ! $h1) {
            if ($delete_initial_heading) {
                s|(<p>)?\s*$stag\Q$h2_full\E$etag(\s*</p>)?(\s*\r?\n)?|$contents|is;
            }
            elsif ($contents) {
                s|((<p>)?\s*$stag\Q$h2_full\E$etag(\s*</p>)?(\s*\r?\n)?)|$1$contents|is;
            }
        }
    }

    return wantarray ? ($h1, $h2, $_) : $h1
}

#
# Additional html-specific markup (heading anchors, heading parsing, 
#   pseudotags).
#
sub markup_html2
{
    my %arg = @_;
    local $_ = $arg{input};
    my $markup_heading_anchors = $arg{markup_heading_anchors};
    $markup_heading_anchors = 1 if ! defined $markup_heading_anchors;
    my $markup_contents = $arg{markup_contents};
    $markup_contents = 1 if ! defined $markup_contents;

    # Markup named anchors
    # Remove spaces, punctuation, tags and entities from labels
    while (s|("[^"]*"=#\S*)[ \t]+|$1_|g) { 1; }               #"
    while (s|("[^"]*"=#\w*)</?[^h>/][^>]*>|$1_|ig) { 1; }        #"
    while (s|("[^"]*"=#\w*)&[^;]{2,6};|$1_|g) { 1; }          #"
    while (s|("[^"]*"=#\w*)[-!?"'/\\:;,.()@#\$%^&*[\]]+|$1_|g) { 1; }     #'
    while (s|("[^"]*"=#\w*?)__+|$1_|g) { 1; }                 #"
    # Remove leading and trailing underscores from labels
    s|("[^"]*"=#)_|$1|g;                                      #"
    s|("[^"]*"=#\w+)_(\W)|$1$2|g;                             #"
    # Markup labelled and unlabelled named anchors
    s|"([^"]*)"=#(\w[^ \t<]*)|<a name="$2">$1</a>|mg;       #"
    s|"([^"]*)"=#([^ \t<]+)|<a name="ref$2">$1</a>|mg;      #"
    s|^=#(\w[^ \t<]*)|<a name="$1"></a>|mg;
    s|^=#([^ \t<]+)|<a name="ref$1"></a>|mg;

    # Parse for top-level headings
    if ($arg{delete_initial_heading} || 
            ref $arg{h1} eq 'SCALAR' || 
            ref $arg{h2} eq 'SCALAR' || 
            $arg{auto_contents}) {
        my ($h1, $h2);
        ($h1, $h2, $_) = parse_headings(@_, input => $_);
        ${$arg{h1}} = $h1 if $h1 && ref $arg{h1} eq 'SCALAR';
        ${$arg{h2}} = $h2 if $h2 && ref $arg{h2} eq 'SCALAR';
    }

    # Replace <contents> pseudotag
    my $contents_tag;
    if ($markup_contents && 
            (($contents_tag) = m!((<|&lt;)contents\b(.*?)(>|&gt;))!is)) {
        my $attr = $3;
        my $contents = '';
        my $type = 'ul';
        my $indent = 2;
        if ($attr =~ m/\btype="?(\w+)"?\b/i) {
            $type = $1;
        }
        if ($attr =~ m/\bindent="?(\d+)"?\b/i) {
            $indent = $1;
        }
        {
            # Ignore everything up to the contents tag
            local $_ = $_;
            s/^.*$contents_tag//s;
            my @headings = m!(<(h[1-6])\b.*?\</\2>)!ig;
            if (@headings) {
                my %indent = ();
                if ($indent) {
                    # Cache the heading levels found in %h
                    for my $h (@headings) {
                        $indent{$1} = 1 if ($h =~ m|<h([1-6])\b|);
                    }
                    my $i = 0;
                    for my $j (1 .. 6) {
                        $indent{$j} = $i++ if $indent{$j};
                    }
                }
                else {
                    $contents .= "<$type>\n" if $type ne 'simple';
                }
                $contents .= "<p>" if $type eq 'simple';
                my $lasthead = 0;
                for my $h (@headings) {
                    # Pick out named anchor label (if any) and heading
#                   if ($h =~ m|((?:<a name="([^"]+))?.*)?<h([1-6])\b[^>]*>(.+?)</h\3>|i) {
                    if ($h =~ m|<h([1-6])\b[^>]*>(<a name="([^"]+)">)?(.+?)(</a>)?</h\1>|i) {
                        my $head = $1;
                        if ($type eq 'simple') {
                            $contents .= '&nbsp;' x ($indent * $indent{$head});
                        }
                        else {
                            if ($head > $lasthead) {
                                $contents .= "<$type>";
                            }
                            elsif ($head < $lasthead) {
                                $contents .= "</$type>";
                            }
                            $contents .= '<li>';
                        }
                        if ($3) {
                            $contents .= qq(<a href="#$3">$4</a>);
                        } 
                        else {
                            $contents .= $4;
                        }
                        $contents .= ($type ne 'simple') ? "</li>\n" : "<br>\n";
                        $lasthead = $head;
                    }
                }
                if ($type eq 'simple') {
                    $contents .= "</p>";
                }
                else {
                    for (0 .. $indent{$lasthead}) {
                        $contents .= "</$type>\n";
                    }
                }
            }
        }
        # Escape regex special chars
        $contents_tag =~ s/([\\^\.\$\|\(\)\[\]])/\\$1/g;
        s|$contents_tag|$contents|i;
    }

    return $_;
}

#
# Markup the given content.
#
sub markup
{
    my %arg = @_;
    my $input = $arg{input};
    my $output = $arg{output};

    croak "[markup] no content found to markup" if ! $input;
    croak "[markup] 'output' arg not reference" 
        if $output && ! ref $output eq 'SCALAR';

    # Markup text -> html
    my $html = markup_html1(@_);
    $html = markup_text(@_, input => $html) 
        if $html !~ m/^[^<]*<(html|!doctype)/i;
    # Markup html
    $html = markup_html2(@_, input => $html);

    $$output = $html if ref $output;
    return $html;
}

1;

__END__

