# tag: HTML Breadcrumb class

package OF::Breadcrumbs;

use File::Basename;
use Carp;
use Data::Dumper;
use strict;
require Exporter;

our $VERSION = '0.05';
our @ISA = qw(Exporter);
our @EXPORT = ();
our @EXPORT_OK = qw(breadcrumbs);
our %EXPORT_TAGS = (
    all => [ @EXPORT_OK ],
);

# my @ARG = qw(path roots indexes map omit labels apache_req sep format format_last);
my @ARG = qw(path roots indexes labels sep format format_last extra);

sub init
{
    my $self = shift;
    my %arg = (
        sep => '&nbsp;&gt;&nbsp;',
        indexes => [ 'index.html' ],
        format => '<a href="%s">%s</a>',
        format_last => '%s',
        @_,
    );

    # Check for invalid args
    my %ARG = map { $_ => 1 } @ARG;
    my @bad = grep { ! exists $ARG{$_} } keys %arg;
    croak "[Breadcrumbs::init] invalid argument(s): " . join(',',@bad) if @bad;

    # Add arguments to $self
    @$self{ @ARG } = @arg{ @ARG };

    return $self;
}

sub new
{
    my $class = shift;
    my $self = {};
    bless $self, $class;
    return $self->init(@_);
}

#
# Split the path into elements (stored in $self->{elt} arrayref)
#
sub split
{
    my $self = shift;

    # Identify the root
    $self->{elt} = [];
    $self->{roots} ||= [];
    $self->{roots} = [ $self->{roots} ] if $self->{roots} && ! ref $self->{roots};
    my $root = '/';
    for my $r (@{$self->{roots}}) {
        if ($self->{path} =~ m/^$r\b/) {
            $root = $r;
            $root .= '/' if substr($root,-1) ne '/';
            last;
        }
    }
    push @{$self->{elt}}, $root;

    # Add elements
    my $current = $root;
    while ($self->{path} =~ m|^$current.*?([^/]+/?)|) {
        $current .= $1;
        push @{$self->{elt}}, $current;
    }

    # Check the final element for indexes
    $self->{indexes} = [ $self->{indexes} ] 
        if $self->{indexes} && ! ref $self->{indexes};
    if (ref $self->{indexes} eq 'ARRAY') {
        my $final = basename($self->{elt}->[ $#{$self->{elt}} ]);
        if (grep /^$final$/, @{$self->{indexes}}) {
            pop @{$self->{elt}};
        }
    }
}

#
# Generate a default label for $elt
#
sub label_default
{
    my $self = shift;
    my ($elt, $extra) = @_;
    my $label = '';

    if ($elt eq '/' || $elt eq '') {
        $label = 'Home';
    }
    else {
        $elt = substr($elt,0,-1) if substr($elt,-1) eq '/';
        $label = basename($elt);
        $label = ucfirst($label) if lc($label) eq $label && $label =~ m/^\w+$/;
    }

    return $label;
}

#
# Return a label for the given element
#
sub label
{
    my $self = shift;
    my ($elt, $extra) = @_;
    my $label = '';

    # Check $self->{labels}
    if (ref $self->{labels} eq 'CODE') {
        my $base = basename(substr($elt,-1) eq '/' ? substr($elt,0,-1) : $elt);
        $label = $self->{labels}->($elt, $base, $extra);
    }
    elsif (ref $self->{labels} eq 'HASH') {
        $elt = substr($elt,0,-1) if substr($elt,-1) eq '/';
        $label ||= $self->{labels}->{$elt};
        $label ||= $self->{labels}->{$elt . '/'};
        $label ||= $self->{labels}->{basename($elt)};
        $label ||= $self->{labels}->{'/' . basename($elt)};
        $label ||= $self->{labels}->{'/' . basename($elt) . '/'};
    }

    # Else use defaults
    $label ||= $self->label_default($elt, $extra);

    return $label;
}

# 
# HTML-format the breadcrumbs
#
sub format 
{
    my $self = shift;

    my $out;
    for (my $i = 0; $i <= $#{$self->{elt}}; $i++) {
        my $label = $self->label($self->{elt}->[$i], $self->{extra});

        # Format breadcrumb links
        if ($i != $#{$self->{elt}}) {
            # $self->{format} coderef
            if (ref $self->{format} eq 'CODE') {
                $out .= $self->{format}->($self->{elt}->[$i], $label, 
                    $self->{extra});
            }
            # $self->{format} sprintf pattern
            elsif ($self->{format} && ! ref $self->{format}) {
                $out .= sprintf $self->{format}, $self->{elt}->[$i], $label;
            }
            # Else croak
            else {
                croak "[Breadcrumbs::format] invalid format $self->{format}";
            }

            # Separator
            $out .= $self->{sep};
        }

        # Format final element breadcrumb label
        else {
            # $self->{format_last} coderef
            if (ref $self->{format_last} eq 'CODE') {
                $out .= $self->{format_last}->($label, $self->{extra});
            }
            # $self->{format_last} sprintf pattern
            elsif ($self->{format_last} && ! ref $self->{format_last}) {
                $out .= sprintf $self->{format_last}, $label;
            }
            # Else croak
            else {
                croak "[Breadcrumbs::format] invalid format_last $self->{format_last}";
            }
        }
    }

    return $out;
}

#
# The real work - process and render the given path
#
sub render
{
    my $self = shift;
    my %arg = @_;

    # Check for invalid args
    my %ARG = map { $_ => 1 } @ARG;
    my @bad = grep { ! exists $ARG{$_} } keys %arg;
    croak "[Breadcrumbs::render] invalid argument(s): " . join(',',@bad) if @bad;

    # Add args to $self
    for (@ARG) {
        $self->{$_} = $arg{$_} if defined $arg{$_};
    }

    # Croak if no path
    croak "[Breadcrumbs::render] no 'path' argument found" if ! $self->{path};

    # Split the path into elements
    $self->split();

    # Format
    return $self->format();
}

#
# Procedural interface
#
sub breadcrumbs
{
    my $bc = OF::Breadcrumbs->new(@_);
    croak "[breadcrumbs] object creation failed!" if ! ref $bc;
    return $bc->render();
}

1;

__END__

=head1 NAME

OF::Breadcrumbs - module to produce HTML 'breadcrumb trails'.


=head1 SYNOPSIS

    # Procedural interace
    use OF::Breadcrumbs qw(breadcrumbs);
    print breadcrumbs(path => $path, labels => {
        'download.html' => 'Download',
        foo => 'Bar',
        'x.html' => 'The X Files',
    });

    # Object interface
    use OF::Breadcrumbs;

    # Create
    $bc = OF::Breadcrumbs->new(path => $path, labels => {
        'download.html' => 'Download',
        foo => 'Bar',
        'x.html' => 'The X Files',
    });

    # Render
    print $bc->render(sep => '&nbsp;::&nbsp;');


=head1 DESCRIPTION

OF::Breadcrumbs is a module used to create HTML 'breadcrumb trails'
i.e. an ordered set of html links locating the current page within
a hierarchy. 

OF::Breadcrumbs essentially splits the given path up into a list of
elements, derives labels to use each of these elements, and then 
renders this list as a set of links for each element with the derived 
label.

Both procedural and object-oriented interfaces are provided. The OO 
interface is primarily useful if you want to separate object creation
from rendering (e.g. initialise the object in an initialisation step,
and rendering where the breadcrumbs should appear), or for 
subclassing.

=head2 PROCEDURAL INTERFACE

The procedural interface is the breadcrumbs() subroutine (not
exported by default). Usage:

    # Procedural interace
    use OF::Breadcrumbs qw(breadcrumbs);
    print breadcrumbs(
        path => $path, 
        labels => {
            'download.html' => 'Download',
            foo => 'Bar',
            'x.html' => 'The X Files',
        },
        roots => [ qw(/web/of /home) ],
    );


breadcrumbs() takes the following parameters:

PATH PROCESSING

=over 4

=item *

L<path|path> - the web-relative path of the item this breadcrumb trail 
is for, as found, for example, in $ENV{SCRIPT_NAME}. This should be 
the I<real> path to the object, so that the breadcrumb links work 
correctly - if you want to modify the elements used from the path in 
the breadcrumb trail, see the L<root>, L<map>, and L<omit> parameters.

=item *

L<roots|roots> - an arrayref of fully-qualified paths used to identify
the root (the first element) of the breadcrumb trail as something other 
than '/'. For example, if the roots arrayref contains '/foo', a path of 
/foo/test.html will be split into two elements: /foo and /foo/test/html. 
The default behaviour would be to split /foo/test.html into three 
elements: /, /foo, and /foo/test.html.

=item *

L<indexes|indexes> - an arrayref of filenames (basenames) to treat 
as index pages (defaults to [ 'index.html' ]). Index pages are omitted 
where they occur as the last element in the element list, essentially
identifying the index page with its directory. Anything you add to
an apache DirectoryIndex directive should probably also be included 
here.

=item *

L<map|map> - not yet implemented.

=item *

L<omit|omit> - not yet implemented.

=back

LABELS

=over 4

=item *

L<labels|labels> - a hashref or a subroutine reference used to derive 
the labels of the breadcrumb trail elements.

If a hashref, both the fully-qualified element name (e.g. /foo/bar/bog) 
and the element basename (e.g. 'bog') are looked up in the hashref (both
with and without a trailing '/'). If found, the corresponding value is
used for the element label.

If this parameter is a subroutine reference, the subroutine is invoked 
for each element, with the fully-qualified element name (e.g.
/foo/bar/bog) and the element basename (e.g. 'bog') passed in as 
parameters.

If no label is found for an element, the default behaviour is to use
the element basename as its label, ucfirst()-ed if lowercase and only 
\w characters (directory names typically ucfirst()-ed, filenames not).

=back

RENDERING

=over 4

=item *

L<sep|sep> - the separator (scalar) used between breadcrumb elements. 
Defaults to '&nbsp;&gt;&nbsp;'.

=item *

L<format|format> - a subroutine reference or a (scalar) sprintf pattern
used to format each breadcrumb element except the last (see 
L<format_last>). 

If a subroutine reference, the subroutine is invoked for each element
with the element and the label as parameters i.e. 
C<&$format($element, $label)>.

If a scalar, it is used as a sprintf format with the element and the 
label as arguments i.e. C<sprintf $format, $element, $label>.

format defaults to '<a href="%s">%s</a>' i.e. a vanilla HTML link.

=item *

L<format_last|format_last> - a subroutine reference or a (scalar) 
sprintf pattern used to format the last breadcrumb element (not a 
link).

If a subroutine reference, the subroutine is invoked for the element
the label as only parameter i.e. C<&$format_last($label)>.

If a scalar, it is used as a sprintf format with the label as 
argument i.e. C<sprintf $format_last, $label>.

format_last defaults to '%s' i.e. just the label itself.

=back


=head1 BUGS

None known.


=head1 AUTHOR

Gavin Carr (gavin@openfusion.com.au)


=head1 COPYRIGHT

Copyright 2001-2002, Open Fusion Pty Ltd. All Rights Reserved.

This program is free software. You may copy or redistribute it under the 
same terms as perl itself.

=cut
