# tag: Template base class

package OF::TemplateBase;

use Carp;
use strict;

sub init
{
    my $self = shift;
    my %arg = @_;
    $self->{map} = $arg{map};
    $self->{prefix} = $arg{prefix} if exists $arg{prefix};
    $self->{open} = $arg{open} if exists $arg{open};
    $self->{close} = $arg{close} if exists $arg{close};
    die "missing or invalid map" 
        if ref $self->{map} ne 'HASH';
    die "either prefix or open parameter required" 
      if ! $self->{prefix} && ! $self->{open};

    return $self;
}

sub new
{
    my $class = shift;
    my $self = {};
    bless $self, $class;
    return $self->init(@_);
}

sub munge
{
    my $self = shift;
    my %arg = (autoformat => 1, @_);
    my $text = $arg{text};
    my $infile = $arg{infile};
    my $outfile = $arg{outfile};
    my $autoformat = $arg{autoformat};
    delete $arg{text};
    delete $arg{infile};
    delete $arg{outfile};
    delete $arg{autoformat};
    croak "[OF::TemplateBase] no template found" if ! $text && ! $infile;

    # Make a local copy of the map
    my $map = { %{$self->{map}} };
    # Invoke all coderefs with the remaining args
    for (keys %$map) {
        $map->{$_} =  $map->{$_}->(%arg) if ref $map->{$_} eq 'CODE';
    }

    # Read the template file
    if (! $text && $infile) {
        open INFILE, "<", $infile
            or die "unable to open infile '$infile': $!";
        { 
            local $/ = undef;
            $text = <INFILE>;
        }
        close INFILE;
    }

    # Escape regex chars in open and close
    # Do the actual munge
    my $pattern = '('. $self->{open} . '(' . $self->{prefix} . '\w+)' . $self->{close} . ')';
    $text =~ s/$pattern/$map->{$2} || $1/egs;

    # TODO: Autoformat

    # Write to outfile
    if ($outfile) {
        open OUTFILE, ">", $outfile.$$ 
            or die "unable to open outfile '$outfile.$$': $!";
        print OUTFILE $text;
        close OUTFILE;
        rename $outfile.$$, $outfile
            or die "outfile rename ($outfile.$$ -> $outfile) failed: $!";
    }

    return $text;
}

1;

__END__

=head1 NAME

OF::TemplateBase - Templating base class


=head1 SYNOPSIS

    # Subclass as follows
    package Foo::Template;

    use OF::TemplateBase;
    @ISA = qw(OF::TemplateBase);

    sub init {
        my $map = {
            FOO_HOME => '/opt/foo',
            FOO_VERSION => '1.23',
            APACHE_CONF => '/usr/local/apache/conf',
        };
        $self->SUPER::init(open => '<', close => '>', map => $map);
    }

    1;

    # Invocation (of the subclass)
    my $t = Foo::Template->new();
    $out = $t->munge(template_infile => 'Config.pm.dist', template


=head1 DESCRIPTION

OF::TemplateBase is a simple templating base class used to replace arbitrary
text in template files. It is typically used by a subclass setting up an
application-specific text => value mappings which TemplateBase then uses
when munging the templates.


=head1 BUGS

None known.


=head1 AUTHOR

Gavin Carr (gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2001-2002, Open Fusion Pty. Ltd. All Rights Reserved.

This program is free software. You may copy or redistribute it under the 
same terms as perl itself.

=cut
